import collections

StorageUnit = collections.namedtuple("StorageUnit", ["B", "KiB", "MiB", "GiB", "TiB", "PiB"])

diskunit = StorageUnit(B=1, KiB=1024**1, MiB=1024**2, GiB=1024**3, TiB=1024**4, PiB=1024**5)



import mimetypes
import os
from typing import List, Union
from zipfile import ZipFile

import click
import igittigitt


def zip_files_in_dir(directory_path: str, zip_file: ZipFile, ignore_file: Union[str, None] = None) -> None:
    # Zip the files that matches the filter from given directory
    files = get_files_in_dir(directory_path=directory_path, ignore_file=ignore_file)
    # Iterate over all the files and zip them
    for file in sorted(files):
        zip_file.write(file)


def get_files_in_dir(directory_path: str, ignore_file: Union[str, None] = None) -> set:
    file_list = set()

    ignore_parser = igittigitt.IgnoreParser()
    if ignore_file:
        ignore_parser.parse_rule_files(os.path.dirname(ignore_file), os.path.basename(ignore_file))

    # Iterate over all the files in directory
    for root, _, files in os.walk(directory_path):
        for file in files:
            # Create complete filepath of file in directory
            file_path = os.path.join(root, file)
            if file_path.startswith("./"):
                file_path = file_path[2:]

            # can we add this file to the collection?
            if not ignore_parser.match(file_path):
                file_list.add(file_path)

    return file_list


def zip_paths(paths: List, zip_file: ZipFile, exclude_paths: List = []) -> None:
    files = get_files_in_paths(paths, exclude_paths)

    # Iterate over all the files and zip them
    for file in sorted(files):
        zip_file.write(file)


def get_files_in_paths(paths: List, exclude_paths: List = []) -> list:
    files = set()

    # first filter out empty string paths
    for path in filter(lambda x: len(x), paths):
        # replace asteriks if possible
        if path.startswith("*"):
            path = "." + path[1:]

        if path.endswith("/*"):
            path = path[:-1]

        if not os.path.exists(path):
            click.echo(f"{path} does not exists...skipping")
            continue

        # base folder path and check whether this is in the exclude_list
        base_folder = "/".join(path.split("/")[:2])
        if base_folder in exclude_paths:
            click.echo(
                f"[CONFIG ERROR] '{base_folder}' from '{path}' is on the exclusion list. AskAnna does not allow to "
                "access these files. This path is ignored and we continue with the other paths.",
                err=True,
            )
            continue

        if os.path.isdir(path):
            files.update(get_files_in_dir(path))
        else:  # so, we got a file
            if path.startswith("./"):
                path = path[2:]
            files.add(path)

    return sorted(files)


def create_zip_from_paths(filename: str, paths: List = []) -> None:
    """
    Create a ZipFile on the given `filename` location
    """
    # We exclude the following directories from included into the zip
    exclude_paths = [
        "/",
        "/bin",
        "/dev",
        "/lib",
        "/mnt",
        "/opt",
        "/proc",
        "/usr",
        "/var",
    ]

    with ZipFile(filename, mode="w") as f:
        zip_paths(paths, f, exclude_paths=exclude_paths)


def file_type(path):
    """Mimic the type parameter of a JS File object.
    Resumable.js uses the File object's type attribute to guess mime type,
    which is guessed from file extention accoring to
    https://developer.mozilla.org/en-US/docs/Web/API/File/type.
    Parameters
    ----------
    path : str
        The path to guess the mime type of
    Returns
    -------
    str
        The inferred mime type, or '' if none could be inferred
    """
    type_, _ = mimetypes.guess_type(path)
    # When no type can be inferred, File.type returns an empty string
    return "" if type_ is None else type_


def content_type_file_extension(content_type: str) -> str:
    content_type_file_extension_mapping = {
        "application/csv": ".csv",
        "application/json": ".json",
        "application/pdf": ".pdf",
        "application/vnd.ms-excel": ".xls",
        "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet": ".xlsx",
        "application/zip": ".zip",
        "image/jpeg": ".jpeg",
        "image/png": ".png",
        "text/plain": ".txt",
        "text/xml": ".xml",
    }

    file_extension = content_type_file_extension_mapping.get(content_type, ".unknown")

    return file_extension



import datetime
from dataclasses import dataclass
from typing import Dict

from dateutil import parser as dateutil_parser

from .relation import ProjectRelation, WorkspaceRelation


@dataclass
class Job:
    suuid: str
    name: str
    description: str
    environment: str
    timezone: str
    schedules: list
    notifications: dict
    project: ProjectRelation
    workspace: WorkspaceRelation
    created_at: datetime.datetime
    modified_at: datetime.datetime

    @classmethod
    def from_dict(cls, data: Dict) -> "Job":
        data["created_at"] = dateutil_parser.parse(data["created_at"])
        data["modified_at"] = dateutil_parser.parse(data["modified_at"])

        project = ProjectRelation.from_dict(data["project"])
        del data["project"]
        workspace = WorkspaceRelation.from_dict(data["workspace"])
        del data["workspace"]

        return cls(project=project, workspace=workspace, **data)


@dataclass
class Payload:
    suuid: str
    size: int
    lines: int
    created_at: datetime.datetime
    modified_at: datetime.datetime

    def __str__(self):
        return (
            f"Payload: {self.suuid} ({self.size} byte"
            + ("s" if self.size != 1 else "")
            + f" & {self.lines} line"
            + ("s" if self.lines != 1 else "")
            + ")"
        )

    def __repr__(self):
        return f"Payload(suuid='{self.suuid}', size={self.size}, lines={self.lines})"

    @classmethod
    def from_dict(cls, data: Dict) -> "Payload":
        data["created_at"] = dateutil_parser.parse(data["created_at"])
        data["modified_at"] = dateutil_parser.parse(data["modified_at"])
        return cls(**data)



import io
import os
from typing import Optional

import resumable

from askanna.core.exceptions import PostError
from askanna.core.utils.file import file_type
from askanna.core.utils.settings import diskunit
from askanna.gateways.api_client import client


class Upload:
    message_upload_success = "File is uploaded"
    message_upload_fail = "File upload failed"

    def __init__(self):
        self.suuid = None
        self.resumable_file = None

    def register_upload_url(self) -> str:
        raise NotImplementedError(f"Please implement 'register_upload_url' for {self.__class__.__name__}")

    def register_chunk_url(self) -> str:
        raise NotImplementedError(f"Please implement 'register_chunk_url' for {self.__class__.__name__}")

    def upload_chunk_url(self, chunk_uuid: str) -> str:
        raise NotImplementedError(f"Please implement 'upload_chunk_url' for {self.__class__.__name__}")

    def finish_upload_url(self) -> str:
        raise NotImplementedError(f"Please implement 'finish_upload_url' for {self.__class__.__name__}")

    @property
    def chunk_baseinfo(self) -> dict:
        return {
            "resumableChunkSize": self.resumable_file.chunk_size,  # type: ignore
            "resumableTotalSize": self.resumable_file.size,  # type: ignore
            "resumableType": file_type(self.resumable_file.path),  # type: ignore
            "resumableIdentifier": str(self.resumable_file.unique_identifier),  # type: ignore
            "resumableFilename": os.path.basename(self.resumable_file.path),  # type: ignore
            "resumableRelativePath": self.resumable_file.path,  # type: ignore
            "resumableTotalChunks": len(self.resumable_file.chunks),  # type: ignore
            "resumableChunkNumber": 1,
            "resumableCurrentChunkSize": 1,
        }.copy()

    def upload(self, file_path: str):
        self.suuid = self.create_entry(file_path)
        self.upload_file(file_path)
        return self.finish_upload()

    @property
    def entry_extrafields(self) -> dict:
        return {}

    def create_entry(self, file_path) -> str:
        request_dict = {
            "filename": os.path.basename(file_path),
            "size": os.stat(file_path).st_size,
        }

        request_dict.update(**self.entry_extrafields)

        # Register upload on the server
        reg_upload = client.post(self.register_upload_url(), json=request_dict)
        if reg_upload.status_code != 201:
            raise PostError(
                f"In the AskAnna platform something went wrong with creating the upload entry: {reg_upload.json()}",
            )

        return reg_upload.json().get("suuid")

    def upload_file(self, file_path):
        """
        Take a file and make chunks out of it to upload
        """
        self.resumable_file = resumable.file.ResumableFile(file_path, 1 * diskunit.MiB)  # type: ignore
        for chunk in self.resumable_file.chunks:
            self.upload_chunk(chunk)

    @property
    def chunk_dict_template(self) -> dict:
        return {
            "filename": "",
            "size": 0,
            "file_no": 0,
            "is_last": False,
        }

    def upload_chunk(self, chunk):
        chunk_dict = self.chunk_dict_template.copy()
        chunk_dict.update(
            **{
                "filename": chunk.index + 1,
                "size": chunk.size,
                "file_no": chunk.index + 1,
                "is_last": len(self.resumable_file.chunks) == chunk.index + 1,  # type: ignore
            }
        )

        # Register chunk on the server
        reg_chunk = client.post(self.register_chunk_url(), json=chunk_dict)
        if reg_chunk.status_code != 201:
            raise PostError(
                "In the AskAnna platform something went wrong with creating the chunk entry for file_no "
                f"'{chunk_dict.get('file_no')}'.",
            )
        chunk_uuid = reg_chunk.json().get("uuid")

        files = {"file": io.BytesIO(chunk.read())}
        data = self.chunk_baseinfo
        data.update(
            **{
                "resumableChunkNumber": chunk.index + 1,
                "resumableCurrentChunkSize": chunk.size,
            }
        )

        upload_chunk_response = client.post(
            self.upload_chunk_url(chunk_uuid=chunk_uuid),
            data=data,
            files=files,
        )

        if upload_chunk_response.status_code != 200:
            raise PostError(
                f"Chunk with file_no '{chunk_dict.get('file_no')}' and chunk UUID '{chunk_uuid}' could not be uploaded"
            )

    def finish_upload(self):
        # Do final call when all chunks are uploaded
        final_call_dict = self.chunk_baseinfo
        final_call_req = client.post(self.finish_upload_url(), data=final_call_dict)

        if final_call_req.status_code == 200:
            return True, self.message_upload_success
        else:
            return False, self.message_upload_fail


class PackageUpload(Upload):
    message_upload_success = "Package is uploaded"
    message_upload_fail = "Package upload failed"

    def __init__(self, project_suuid: str, description: Optional[str] = None):
        self.project_suuid = project_suuid
        self.description = description
        super().__init__()

    def register_upload_url(self) -> str:
        return client.askanna_url.package.base_package_url

    def register_chunk_url(self) -> str:
        if self.suuid:
            return client.askanna_url.package.package_chunk(self.suuid)
        raise TypeError("No SUUID found for package upload")

    def upload_chunk_url(self, chunk_uuid: str) -> str:
        if self.suuid:
            return client.askanna_url.package.package_chunk_upload(self.suuid, chunk_uuid)
        raise TypeError("No SUUID found for package upload")

    def finish_upload_url(self) -> str:
        if self.suuid:
            return client.askanna_url.package.package_finish_upload(self.suuid)
        raise TypeError("No SUUID found for package upload")

    @property
    def entry_extrafields(self) -> dict:
        return {
            "project_suuid": self.project_suuid,
            "description": self.description,
        }


class ArtifactUpload(Upload):
    message_upload_success = "Artifact is uploaded"
    message_upload_fail = "Artifact upload failed"

    def __init__(self, run_suuid: str):
        self.run_suuid = run_suuid
        super().__init__()

    def register_upload_url(self) -> str:
        if self.run_suuid:
            return client.askanna_url.run.artifact_list(self.run_suuid)
        raise TypeError("No Run SUUID found for artifact upload")

    def register_chunk_url(self) -> str:
        if self.run_suuid and self.suuid:
            return client.askanna_url.run.artifact_chunk(self.run_suuid, self.suuid)
        raise TypeError("No Run SUUID or Artifact SUUID found for artifact upload")

    def upload_chunk_url(self, chunk_uuid: str) -> str:
        if self.run_suuid and self.suuid:
            return client.askanna_url.run.artifact_chunk_upload(self.run_suuid, self.suuid, chunk_uuid)
        raise TypeError("No Run SUUID or Artifact SUUID found for artifact upload")

    def finish_upload_url(self) -> str:
        if self.run_suuid and self.suuid:
            return client.askanna_url.run.artifact_finish_upload(self.run_suuid, self.suuid)
        raise TypeError("No Run SUUID or Artifact SUUID found for artifact upload")


class ResultUpload(Upload):
    message_upload_success = "Result is uploaded"
    message_upload_fail = "Result upload failed"

    def __init__(self, run_suuid: str):
        self.run_suuid = run_suuid
        super().__init__()

    def register_upload_url(self) -> str:
        if self.run_suuid:
            return client.askanna_url.run.result_upload(self.run_suuid)
        raise TypeError("No Run SUUID found for result upload")

    def register_chunk_url(self) -> str:
        if self.run_suuid and self.suuid:
            return client.askanna_url.run.result_chunk(self.run_suuid, self.suuid)
        raise TypeError("No Run SUUID or Result SUUID found for result upload")

    def upload_chunk_url(self, chunk_uuid: str) -> str:
        if self.run_suuid and self.suuid:
            return client.askanna_url.run.result_chunk_upload(self.run_suuid, self.suuid, chunk_uuid)
        raise TypeError("No Run SUUID or Result SUUID found for result upload")

    def finish_upload_url(self) -> str:
        if self.run_suuid and self.suuid:
            return client.askanna_url.run.result_finish_upload(self.run_suuid, self.suuid)
        raise TypeError("No Run SUUID or Result SUUID found for result upload")



from askanna.config import config


def base_url() -> str:
    return config.server.remote + "/v1/"


class AskAnnaURL:
    def __init__(self):
        self.auth = AuthURL()
        self.job = JobURL()
        self.package = PackageURL()
        self.project = ProjectURL()
        self.run = RunURL()
        self.variable = VariableURL()
        self.workspace = WorkspaceURL()

    @property
    def base_url(self) -> str:
        return base_url()


class AuthURL:
    @property
    def base_auth_url(self) -> str:
        return base_url() + "auth/"

    def login(self) -> str:
        return self.base_auth_url + "login/"

    def logout(self) -> str:
        return self.base_auth_url + "logout/"

    def user(self) -> str:
        return self.base_auth_url + "user/"


class JobURL:
    @property
    def base_job_url(self) -> str:
        return base_url() + "job/"

    def job(self) -> str:
        return self.base_job_url

    def job_list(self) -> str:
        return self.base_job_url

    def job_detail(self, job_suuid: str) -> str:
        return self.base_job_url + job_suuid + "/"

    def run_request(self, job_suuid: str) -> str:
        return self.job_detail(job_suuid) + "run/request/batch/"


class PackageURL:
    @property
    def base_package_url(self) -> str:
        return base_url() + "package/"

    def package(self) -> str:
        return self.base_package_url

    def package_list(self) -> str:
        return self.base_package_url

    def package_detail(self, package_suuid: str) -> str:
        return self.package() + package_suuid + "/"

    def package_chunk(self, package_suuid: str) -> str:
        return self.package_detail(package_suuid) + "packagechunk/"

    def package_chunk_upload(self, package_suuid: str, chunk_uuid: str) -> str:
        return self.package_chunk(package_suuid) + chunk_uuid + "/chunk/"

    def package_finish_upload(self, package_suuid: str) -> str:
        return self.package_detail(package_suuid) + "finish_upload/"

    def package_download(self, package_suuid: str) -> str:
        return self.package_detail(package_suuid) + "download/"


class ProjectURL:
    @property
    def base_project_url(self) -> str:
        return base_url() + "project/"

    def project(self) -> str:
        return self.base_project_url

    def project_list(self) -> str:
        return self.base_project_url

    def project_detail(self, project_suuid: str) -> str:
        return self.base_project_url + project_suuid + "/"


class RunURL:
    @property
    def base_run_url(self) -> str:
        return base_url() + "run/"

    def run(self) -> str:
        return self.base_run_url

    def run_list(self) -> str:
        return self.base_run_url

    def run_detail(self, run_suuid: str) -> str:
        return self.base_run_url + run_suuid + "/"

    def payload_list(self, run_suuid: str) -> str:
        return self.run_detail(run_suuid) + "payload/"

    def payload_download(self, run_suuid: str, payload_suuid: str) -> str:
        return self.payload_list(run_suuid) + payload_suuid + "/"

    def status(self, run_suuid: str) -> str:
        return self.run_detail(run_suuid) + "status/"

    def manifest(self, run_suuid: str) -> str:
        return self.run_detail(run_suuid) + "manifest/"

    def metric(self, run_suuid: str) -> str:
        return self.run_detail(run_suuid) + "metric/"

    def metric_detail(self, run_suuid: str) -> str:
        return self.metric(run_suuid) + run_suuid + "/"

    def variable(self, run_suuid: str) -> str:
        return self.run_detail(run_suuid) + "variable/"

    def variable_detail(self, run_suuid: str) -> str:
        return self.variable(run_suuid) + run_suuid + "/"

    def log(self, run_suuid: str) -> str:
        return self.run_detail(run_suuid) + "log/"

    def artifact(self, run_suuid: str) -> str:
        return self.run_detail(run_suuid) + "artifact/"

    def artifact_list(self, run_suuid: str) -> str:
        return self.run_detail(run_suuid) + "artifact/"

    def artifact_detail(self, run_suuid: str, artifact_suuid: str) -> str:
        return self.run_detail(run_suuid) + "artifact/" + artifact_suuid + "/"

    def artifact_chunk(self, run_suuid, artifact_suuid: str) -> str:
        return self.artifact_detail(run_suuid, artifact_suuid) + "artifactchunk/"

    def artifact_chunk_upload(self, run_suuid: str, artifact_suuid: str, chunk_uuid: str) -> str:
        return self.artifact_chunk(run_suuid, artifact_suuid) + chunk_uuid + "/chunk/"

    def artifact_finish_upload(self, run_suuid: str, artifact_suuid: str) -> str:
        return self.artifact_detail(run_suuid, artifact_suuid) + "finish_upload/"

    def artifact_download(self, run_suuid: str, artifact_suuid: str) -> str:
        return self.artifact_detail(run_suuid, artifact_suuid) + "download/"

    def result(self, run_suuid: str) -> str:
        return self.run_detail(run_suuid) + "result/"

    def result_upload(self, run_suuid: str) -> str:
        return self.run_detail(run_suuid) + "result-upload/"

    def result_upload_detail(self, run_suuid: str, result_suuid: str) -> str:
        return self.result_upload(run_suuid) + result_suuid + "/"

    def result_chunk(self, run_suuid, result_suuid: str) -> str:
        return self.result_upload_detail(run_suuid, result_suuid) + "resultchunk/"

    def result_chunk_upload(self, run_suuid: str, result_suuid: str, chunk_uuid: str) -> str:
        return self.result_chunk(run_suuid, result_suuid) + chunk_uuid + "/chunk/"

    def result_finish_upload(self, run_suuid: str, result_suuid: str) -> str:
        return self.result_upload_detail(run_suuid, result_suuid) + "finish_upload/"


class VariableURL:
    @property
    def base_variable_url(self) -> str:
        return base_url() + "variable/"

    def variable(self) -> str:
        return self.base_variable_url

    def variable_list(self) -> str:
        return self.base_variable_url

    def variable_detail(self, variable_suuid: str) -> str:
        return self.base_variable_url + variable_suuid + "/"


class WorkspaceURL:
    @property
    def base_workspace_url(self) -> str:
        return base_url() + "workspace/"

    def workspace(self) -> str:
        return self.base_workspace_url

    def workspace_list(self) -> str:
        return self.base_workspace_url

    def workspace_detail(self, workspace_suuid: str) -> str:
        return self.base_workspace_url + workspace_suuid + "/"


askanna_url = AskAnnaURL()



from typing import Optional

__all__ = [
    "ListMixin",
]


class ListMixin:
    """Base class for list method

    This class contains the list method, which is used by all SDKs that have a list method. The list method contains
    logic to get all results from the API, even if the number of results is higher than the page size.

    To use this class, you need to set the gateway attribute to the gateway class

    Example:

        class WorkspaceSDK(ListMixin):
            gateway = WorkspaceGateway()
    """

    gateway = None

    def __init__(self):
        self.list_total_count = None

    def get_gateway(self):
        assert self.gateway is not None, "Gateway is not set"
        return self.gateway

    def list(
        self,
        number_of_results: int = 100,
        order_by: Optional[str] = None,
        other_query_params: Optional[dict] = None,
    ) -> list:
        gateway = self.get_gateway()
        query_params = {
            "page_size": number_of_results,
            "order_by": order_by,
        }
        if other_query_params:
            assert isinstance(other_query_params, dict), "other_query_params must be a dict"
            query_params.update(other_query_params)

        list_response = gateway.list(**query_params)
        results = list_response.results

        self.list_total_count = list_response.total_count
        if number_of_results > len(results):
            while list_response.next_url is not None and len(results) < number_of_results:
                list_response = gateway.list(
                    cursor=list_response.next_url_cursor,
                    **query_params,
                )
                results.extend(list_response.results)

        return results[:number_of_results]



from typing import List, Union

import click
import requests

from askanna import ASKANNA_VERSION
from askanna.core.dataclasses.run import Label
from askanna.core.utils.object import (
    get_type,
    object_fullname,
    prepare_and_validate_value,
)
from askanna.settings import PYPI_PROJECT_URL


def update_available() -> bool:
    """
    Check whether most recent release of AskAnna on PyPI is newer than the AskAnna version in use. If a newer version
    is available, return a info message with update instructions.
    """
    try:
        r = requests.get(PYPI_PROJECT_URL)
    except requests.exceptions.ConnectionError:
        return False
    except requests.exceptions.HTTPError:
        return False
    else:
        if r.status_code == 200:
            pypi_info = r.json()
        else:
            return False

    if ASKANNA_VERSION == pypi_info["info"]["version"]:
        return False
    else:
        click.echo("[INFO] A newer version of AskAnna is available. Update via: pip install -U askanna")
        return True


def make_label_list(label: Union[str, list, dict]) -> List[Label]:
    label_list = []

    if not label:
        return label_list

    # If label is of type string or list, then convert it to tag
    if isinstance(label, str):
        label_list.append(Label(name=label, value=None, type="tag"))
    elif isinstance(label, list):
        for name in label:
            label_list.append(Label(name=name, value=None, type="tag"))
    elif isinstance(label, dict):
        for name, value in label.items():
            if value is None:
                label_list.append(Label(name=name, value=None, type="tag"))
            else:
                value, valid = prepare_and_validate_value(value)
                if valid:
                    label_list.append(Label(name=name, value=value, type=get_type(value)))
                else:
                    click.echo(
                        f"AskAnna cannot store the datatype '{object_fullname(value)}'. Label not stored for '{name}' "
                        f"with value '{value}'.",
                        err=True,
                    )

    return label_list



import os
import sys

import click

from askanna.cli.run_utils.utils import string_expand_variables
from askanna.config import config
from askanna.core.upload import ResultUpload

HELP = """
At the end of a run push the result to AskAnna
"""

SHORT_HELP = "Push result to AskAnna"


@click.command(help=HELP, short_help=SHORT_HELP)
@click.option(
    "--run",
    "run_suuid",
    required=True,
    envvar="AA_RUN_SUUID",
    help="The run SUUID",
)
@click.option(
    "--job-name",
    "job_name",
    required=True,
    envvar="AA_JOB_NAME",
    help="The name of the job",
)
def cli(run_suuid, job_name):
    project_config = config.project.config_dict

    # First check whether we need to create result or not.
    # If output.result is not specified, we skip this step and report this to the stdout.
    result_path = project_config[job_name].get("output", {}).get("result")

    if not result_path:
        click.echo("  Result: no `result` defined for this job in `askanna.yml`")
        sys.exit(0)
    elif isinstance(result_path, list):
        click.echo("  Please enter a path in `result` definition, not a list.", err=True)
        sys.exit(1)

    # Expand and translate result_path if they are configured with variables
    result_path = string_expand_variables([result_path])[0]

    if not os.path.exists(result_path):
        click.echo(f"  Result: {result_path} does not exist. Not saving result.", err=True)
        sys.exit(1)

    click.echo("  Uploading result to AskAnna...")

    try:
        uploader = ResultUpload(run_suuid)
        status, msg = uploader.upload(result_path)
    except Exception as e:
        click.echo(f"  {e}", err=True)
        sys.exit(1)

    if status:
        click.echo(f"  {msg}")
        sys.exit(0)
    else:
        click.echo(f"  {msg}", err=True)
        sys.exit(1)



import sys

import click

from askanna.cli.utils import (
    ask_which_job,
    ask_which_project,
    ask_which_run,
    ask_which_workspace,
    job_run_request,
)
from askanna.config import config
from askanna.config.utils import string_format_datetime
from askanna.core.dataclasses.run import STATUS
from askanna.core.exceptions import GetError, PatchError
from askanna.sdk.run import RunSDK

try:
    from typing import get_args
except ImportError:  # pragma: no cover
    from typing_extensions import get_args

HELP = """
This command will allow you to start a run in AskAnna.
"""

SHORT_HELP = "Start a run in AskAnna"


class SkipArgForSubCommand(click.Group):
    def parse_args(self, ctx, args):
        if args and args[0] in self.commands:
            if len(args) == 1 or args[1] not in self.commands:
                args.insert(0, "")
        super().parse_args(ctx, args)


@click.group(help=HELP, short_help=SHORT_HELP, invoke_without_command=True, cls=SkipArgForSubCommand)
@click.argument("job_name", required=False, type=str)
@click.option("--id", "-i", "job_suuid", required=False, type=str, help="SUUID of the job to run")
@click.option("--data", "-d", required=False, type=str, default=None, help="JSON data")
@click.option(
    "--data-file",
    "-D",
    "data_file",
    required=False,
    type=str,
    default=None,
    help="File with JSON data",
)
@click.option(
    "--push/--no-push",
    "-p",
    "push_code",
    default=False,
    show_default=True,
    help="Push code before starting a run",
)
@click.option("--name", "-n", required=False, type=str, help="Give the run a name")
@click.option(
    "--description",
    required=False,
    type=str,
    help="Description of the run",
    default="",
)
@click.option("--project", "project_suuid", required=False, type=str, help="Project SUUID")
@click.option("--workspace", "workspace_suuid", required=False, type=str, help="Workspace SUUID")
@click.pass_context
def cli(
    ctx,
    job_name,
    job_suuid,
    name,
    description,
    data,
    data_file,
    push_code,
    project_suuid,
    workspace_suuid,
):
    if ctx.invoked_subcommand is None:
        job_run_request(
            job_name=job_name,
            job_suuid=job_suuid,
            name=name,
            description=description,
            data=data,
            data_file=data_file,
            push_code=push_code,
            project_suuid=project_suuid,
            workspace_suuid=workspace_suuid,
        )


@cli.command(help="List runs available in AskAnna", short_help="List runs")
@click.option(
    "--status",
    "status",
    required=False,
    type=click.Choice(list(get_args(STATUS)), case_sensitive=False),
    help="Show runs with a specific run status",
)
@click.option(
    "--job",
    "-j",
    "job_suuid",
    required=False,
    type=str,
    help="Job SUUID to list runs for a job",
)
@click.option(
    "--project",
    "-p",
    "project_suuid",
    required=False,
    type=str,
    help="Project SUUID to list runs for a project",
)
@click.option(
    "--workspace",
    "-w",
    "workspace_suuid",
    required=False,
    type=str,
    help="Workspace SUUID to list runs for a workspace",
)
@click.option("--search", "-s", required=False, type=str, help="Search for a specific run")
def list(status, job_suuid, project_suuid, workspace_suuid, search):
    run_sdk = RunSDK()
    try:
        runs = run_sdk.list(
            number_of_results=100,
            status=status,
            job_suuid=job_suuid,
            project_suuid=project_suuid,
            workspace_suuid=workspace_suuid,
            search=search,
            order_by="job.name,name",
        )
    except Exception as e:
        click.echo(f"Something went wrong while listing the runs:\n  {e}", err=True)
        sys.exit(1)

    if not runs:
        click.echo("We cannot find any run.")
        sys.exit(0)

    if job_suuid:
        click.echo(f"The runs for job '{runs[0].job.name}' are:\n")
        click.echo("")
        click.echo("-------------------    -------------------------")
        click.echo("RUN SUUID              RUN NAME")
        click.echo("-------------------    -------------------------")
    if not job_suuid and project_suuid:
        click.echo(f"The runs for project '{runs[0].project.name}' are:")
    if not job_suuid and not project_suuid and workspace_suuid:
        click.echo(f"The runs for workspace '{runs[0].workspace.name}' are:")
    if not job_suuid:
        click.echo("")
        click.echo("-------------------    --------------------    -------------------    -------------------------")
        click.echo("JOB SUUID              JOB NAME                RUN SUUID              RUN NAME")
        click.echo("-------------------    --------------------    -------------------    -------------------------")

    for run in runs:
        run_name = f"{run.name[:22]}..." if len(run.name) > 25 else run.name[:25]
        if job_suuid:
            click.echo(f"{run.suuid}    {run_name}")
        else:
            job_name = f"{run.job.name[:17]}..." if len(run.job.name) > 20 else run.job.name[:20]
            click.echo(
                "{job_suuid}    {job_name}    {run_suuid}    {run_name}".format(
                    job_suuid=run.job.suuid,
                    job_name=f"{job_name:20}",
                    run_suuid=run.suuid,
                    run_name=run_name,
                )
            )

    if len(runs) != run_sdk.list_total_count:
        click.echo("")
        click.echo(f"Note: the first {len(runs):,} of {run_sdk.list_total_count:,} runs are shown.")

    click.echo("")


@cli.command(help="Get information about a run", short_help="Get run info")
@click.option("--id", "-i", "run_suuid", required=False, type=str, help="Run SUUID")
def info(run_suuid):
    if run_suuid:
        try:
            run = RunSDK().get(run_suuid=run_suuid)
        except GetError as e:
            if str(e).startswith("404"):
                click.echo(f"The run SUUID '{run_suuid}' was not found", err=True)
                sys.exit(1)
            else:
                click.echo(f"Something went wrong while getting info of run SUUID '{run_suuid}':\n  {e}", err=True)
                sys.exit(1)
    else:
        project_suuid = config.project.project_suuid
        if not project_suuid:
            workspace = ask_which_workspace(question="From which workspace do you want to get a run?")
            project = ask_which_project(
                question="From which project do you want to get a run?", workspace_suuid=workspace.suuid
            )
            project_suuid = project.suuid

        job = ask_which_job(question="From which job do you want to get a run?", project_suuid=project_suuid)
        run = ask_which_run(question="Which run do you want to get?", job_suuid=job.suuid)

    if run.metrics_meta.get("count", 0) > 0:
        metric_string = f"{run.metrics_meta.get('count')} metric" + (
            "s" if run.metrics_meta.get("count", 0) > 1 else ""
        )
    else:
        metric_string = "No metrics"

    if run.variables_meta.get("count", 0) > 0:
        variable_string = f"{run.variables_meta.get('count')} variable" + (
            "s" if run.variables_meta.get("count", 0) > 1 else ""
        )
    else:
        variable_string = "No variables"

    print_list = [
        ("Name", run.name),
        ("SUUID", run.suuid),
        ("Description", run.description),
        ("Created by", run.created_by.name),
        None,
        ("Status", run.status),
        ("Duration", f"{run.duration} seconds"),
        ("Started", run.started_at.strftime(string_format_datetime) if run.started_at else "Not started yet"),
        ("Finished", run.finished_at.strftime(string_format_datetime) if run.finished_at else "Not finished yet"),
        None,
        ("Metrics", metric_string),
        ("Variables", variable_string),
        ("Payload", "Yes" if run.payload else "No"),
        ("Result", "Yes" if run.result else "No"),
        ("Artifact", "Yes" if run.artifact else "No"),
        None,
        ("Job", run.job.name),
        ("Job SUUID", run.job.suuid),
        ("Project", run.project.name),
        ("Project SUUID", run.project.suuid),
        ("Workspace", run.workspace.name),
        ("Workspace SUUID", run.workspace.suuid),
        None,
        ("Created", run.created_at.strftime(string_format_datetime)),
        ("Modified", run.modified_at.strftime(string_format_datetime)),
        None,
    ]
    for item in print_list:
        if item is None:
            click.echo("")
        else:
            click.echo(f"{item[0] + ':':16} {item[1]}")


@cli.command(help="Get the status of a run", short_help="Get run status")
@click.option("--id", "-i", "run_suuid", required=False, type=str, help="Run SUUID")
def status(run_suuid):
    run_status = RunSDK().status(run_suuid=run_suuid)
    click.echo(f"Status run SUUID '{run_suuid}': {run_status.status}")


@cli.command(help="Get the log of a run", short_help="Get run log")
@click.option("--id", "-i", "run_suuid", required=False, type=str, help="Run SUUID")
def log(run_suuid):
    run_log = RunSDK().log(run_suuid=run_suuid)
    click.echo(f"Log run SUUID '{run_suuid}':\n")
    for line in run_log:
        click.echo(line[2])


@cli.command(help="Change run information in AskAnna", short_help="Change run")
@click.option("--id", "-i", "run_suuid", required=False, type=str, help="Run SUUID")
@click.option("--name", "-n", required=False, type=str, help="New name to set")
@click.option("--description", "-d", required=False, type=str, help="New description to set")
def change(run_suuid, name, description):
    if not run_suuid:
        project_suuid = config.project.project_suuid
        if not project_suuid:
            workspace = ask_which_workspace(question="From which workspace do you want to change a run?")
            project = ask_which_project(
                question="From which project do you want to change a run?", workspace_suuid=workspace.suuid
            )
            project_suuid = project.suuid

        job = ask_which_job(question="From which job do you want to change a run?", project_suuid=project_suuid)
        run = ask_which_run(question="Which run do you want to change?", job_suuid=job.suuid)
        run_suuid = run.suuid

    if not name and not description:
        if click.confirm("\nDo you want to change the name of the run?"):
            name = click.prompt("New name of the run", type=str)
        if click.confirm("\nDo you want to change the description of the run?"):
            description = click.prompt("New description of the run", type=str)

        click.confirm("\nDo you want to change the run?", abort=True)

    try:
        run = RunSDK().change(run_suuid=run_suuid, name=name, description=description)
    except PatchError as e:
        if str(e).startswith("404"):
            click.echo(f"The run SUUID '{run_suuid}' was not found", err=True)
        else:
            click.echo(f"Something went wrong while changing the run SUUID '{run_suuid}':\n  {e}", err=True)
        sys.exit(1)

    click.echo(f"\nYou succesfully changed run '{run.name}' with SUUID '{run.suuid}'")


@cli.command(help="Remove a run in AskAnna", short_help="Remove run")
@click.option("--id", "-i", "run_suuid", type=str, required=False, help="Run SUUID")
@click.option("--force", "-f", is_flag=True, help="Force")
def remove(run_suuid, force):
    if run_suuid:
        try:
            run = RunSDK().get(run_suuid=run_suuid)
        except GetError as e:
            if str(e).startswith("404"):
                click.echo(f"The run SUUID '{run_suuid}' was not found", err=True)
            else:
                click.echo(f"Something went wrong while getting the info of run SUUID '{run_suuid}':\n  {e}", err=True)
            sys.exit(1)

    else:
        project_suuid = config.project.project_suuid
        if not project_suuid:
            workspace = ask_which_workspace(question="From which workspace do you want to remove a run?")
            project = ask_which_project(
                question="From which project do you want to remove a run?", workspace_suuid=workspace.suuid
            )
            project_suuid = project.suuid

        job = ask_which_job(question="From which job do you want to remove a run?", project_suuid=project_suuid)
        run = ask_which_run(question="Which run do you want to remove?", job_suuid=job.suuid)

    if not force:
        if not click.confirm(f"Are you sure to remove run SUUID '{run.suuid}'?"):
            click.echo("Understood. We are not removing the run.")
            sys.exit(0)

    try:
        _ = RunSDK().delete(run_suuid=run.suuid)
    except Exception as e:
        click.echo(f"Something went wrong while removing the run SUUID '{run.suuid}':\n  {e}", err=True)
        sys.exit(1)

    click.echo(f"You removed run SUUID '{run.suuid}'")



from typing import List, Optional

from askanna.core.dataclasses.variable import Variable
from askanna.gateways.variable import VariableGateway

from .mixins import ListMixin

__all__ = [
    "VariableSDK",
]


class VariableSDK(ListMixin):
    """Management of variables in AskAnna
    This class is a wrapper around the VariableGateway and can be used to manage variables in Python.
    """

    gateway = VariableGateway()

    def list(
        self,
        project_suuid: Optional[str] = None,
        workspace_suuid: Optional[str] = None,
        is_masked: Optional[bool] = None,
        number_of_results: int = 100,
        order_by: Optional[str] = None,
        search: Optional[str] = None,
    ) -> List[Variable]:
        """List variables with filter and order options

        Args:
            project_suuid (str, optional): SUUID of the project to filter on. Defaults to None.
            workspace_suuid (str, optional): SUUID of the workspace to filter on. Defaults to None.
            is_masked (bool, optional): Filter on masked variables. Defaults to None.
            number_of_results (int, optional): Number of results to return. Defaults to 100.
            order_by (str, optional): Order by a specific field. Defaults to None.
            search (str, optional): Search for a specific variable. Defaults to None.

        Raises:
            GetError: Error based on response status code with the error message from the API

        Returns:
            List[Variable]: List of variables. List items are of type Variable dataclass
        """
        return super().list(
            number_of_results=number_of_results,
            order_by=order_by,
            other_query_params={
                "project_suuid": project_suuid,
                "workspace_suuid": workspace_suuid,
                "is_masked": is_masked,
                "search": search,
            },
        )

    def get(self, variable_suuid: str) -> Variable:
        """Get information of a variable

        Args:
            variable_suuid (str): SUUID of the variable

        Raises:
            GetError: Error based on response status code with the error message from the API

        Returns:
            Variable: Variable info in a Variable dataclass
        """
        return self.gateway.detail(variable_suuid=variable_suuid)

    def add(self, project_suuid: str, name: str, value: str, is_masked: bool = False) -> Variable:
        """Add a new variable to a project

        Args:
            project_suuid (str): SUUID of the project
            name (str): Name of the variable
            value (str): Value of the variable
            is_masked (bool, optional): Mask the variable. Defaults to False.

        Raises:
            PostError: Error based on response status code with the error message from the API

        Returns:
            Variable: The newly created variable in a Variable dataclass
        """
        return self.gateway.create(
            project_suuid=project_suuid,
            name=name,
            value=value,
            is_masked=is_masked,
        )

    def change(
        self,
        variable_suuid: str,
        name: Optional[str] = None,
        value: Optional[str] = None,
        is_masked: Optional[bool] = None,
    ) -> Variable:
        """Change the name, value and/or mask of a variable

        Note: is_masked can only be changed to True, a masked variable cannot be set to unmasked.

        Args:
            variable_suuid (str): SUUID of the variable
            name (str, optional): Name of the variable. Defaults to None.
            value (str, optional): Value of the variable. Defaults to None.
            is_masked (bool, optional): Mask the variable. Defaults to None.

        Raises:
            PatchError: Error based on response status code with the error message from the API

        Returns:
            Variable: The updated variable in a Variable dataclass
        """
        return self.gateway.change(
            variable_suuid=variable_suuid,
            name=name,
            value=value,
            is_masked=is_masked,
        )

    def delete(self, variable_suuid: str) -> bool:
        """Delete a variable

        Args:
            variable_suuid (str): SUUID of the variable

        Raises:
            DeleteError: Error based on response status code with the error message from the API

        Returns:
            bool: True if the variable was successfully deleted
        """
        return self.gateway.delete(variable_suuid=variable_suuid)



import sys

import click

from askanna.config import config
from askanna.core.exceptions import GetError
from askanna.core.push import is_project_config_push_ready, push
from askanna.sdk.project import ProjectSDK

HELP = """
Command to make an ZIP archive of the current working folder.
Afterwards we upload (push) this ZIP archive to AskAnna.
"""

SHORT_HELP = "Push code to AskAnna"


@click.command(help=HELP, short_help=SHORT_HELP)
@click.option("--force", "-f", is_flag=True, help="Force push")
@click.option(
    "--description",
    "-d",
    required=False,
    type=str,
    help="Add description to this code",
    default="",
)
@click.option(
    "--message",
    "-m",
    required=False,
    type=str,
    help="Add description to this code",
    default="",
)
def cli(force, description, message):
    if len(description) > 0 and len(message) > 0:
        click.echo("Cannot use both --description and --message.", err=True)
        sys.exit(1)

    if not is_project_config_push_ready():
        sys.exit(1)

    # Check for existing package
    if not force:
        try:
            project = ProjectSDK().get(config.project.project_suuid)
        except GetError:
            click.echo(
                "We cannot find your project on AskAnna. Do you have access to the project with SUUID "
                f"'{config.project.project_suuid}'?",
                err=True,
            )
            sys.exit(1)

        if project.package and not click.confirm("Do you want to replace the current code on AskAnna?"):
            click.echo("We are not pushing your code to AskAnna. You choose to not replace your existing code.")
            sys.exit(0)

    push(overwrite=True, description=description or message)



import math
from pathlib import Path
from typing import Union

import click

from askanna.core.exceptions import GetError
from askanna.core.utils.settings import diskunit
from askanna.gateways.api_client import client


class ChunkedDownload:
    def __init__(self, url: str):
        """
        Takes an url to download from, this can be a url which could redirect to another URI.
        """
        self.history = []
        self.download_queue = []
        self.chunk_size = 10 * diskunit.MiB

        self.url = url
        self.size = 0
        self.accept_ranges = "none"

        self.perform_preflight(url=url)

    @property
    def status_code(self):
        """
        Return the last status_code
        """
        return self.history[-1][1]

    def perform_preflight(self, url):
        """
        We take the session from client and will do a preflight request to
        determine whether the url will result in a (final) http_response_code=200
        """
        response = client.head(url)
        self.history.append([url, response.status_code, response.headers])

        if response.status_code in [301, 302] and response.headers.get("Location"):
            self.url = response.headers.get("Location")
            return self.perform_preflight(url=response.headers.get("Location"))
        if response.status_code == 200:
            self.size = int(response.headers.get("Content-Length", 0))
            self.content_type = response.headers.get("Content-Type")
            self.accept_ranges = response.headers.get("Accept-Ranges", "none")

    def setup_download(self):
        """
        Download the self.url and chunk (if needed) the download
        """
        no_chunks = math.ceil(self.size / self.chunk_size)
        for chunk_no in range(0, no_chunks):
            self.download_queue.append(
                [
                    chunk_no,
                    chunk_no * self.chunk_size,
                    (chunk_no + 1) * self.chunk_size - 1,
                ]
            )
        # Fix last end byte
        self.download_queue[-1][-1] = self.size

    def download(self, output_file: Union[Path, str], overwrite: bool = False):
        """
        Download the target_url, we consume the queue and write partial files.
        In the end we glue them to the original intended file.
        """
        self.setup_download()
        finished_queue = []

        output_file = Path(output_file)
        if output_file.is_dir():
            raise ValueError(f"The output path '{output_file}' is a directory, but should be a file name")
        if output_file.exists() and not overwrite:
            raise ValueError(f"The output file '{output_file}' already exists.")
        output_file.parent.mkdir(parents=True, exist_ok=True)

        chunk_retry = 0
        while len(self.download_queue):
            chunk = self.download_queue.pop(0)
            range_header = {
                "Range": "{accept_ranges}={start}-{end}".format(
                    accept_ranges=self.accept_ranges, start=chunk[1], end=chunk[2]
                )
            }

            try:
                response = client.get(self.url, stream=True, headers=range_header)
                if not response.content or response.status_code not in [200, 206]:
                    raise GetError(
                        f"Could not download chunk {chunk[0]} from {self.url} (code={response.status_code})"
                    )

                # So far so good on the download, save the result
                with Path(f"file_{chunk[0]}.part").open("wb") as f:
                    for r_chunk in response.iter_content(chunk_size=1024):
                        f.write(r_chunk)
            except Exception as e:
                click.echo(e)
                self.download_queue.append(chunk)
                chunk_retry += 1
                if chunk_retry > 5:
                    raise GetError(f"Could not download chunk {chunk[0]} from {self.url} after 5 retries")
            else:
                finished_queue.append(chunk)

        # Combine the chunks in to one file
        with output_file.open("wb") as f:
            finished_queue = sorted(finished_queue, key=lambda x: x[0])
            for chunk in finished_queue:
                chunkfilename = Path(f"file_{chunk[0]}.part")
                with chunkfilename.open("rb") as chunkfile:
                    f.write(chunkfile.read())
                # Delete the chunkfile
                chunkfilename.unlink()



from typing import List, Optional

from askanna.config import config
from askanna.core.dataclasses.job import Job
from askanna.core.dataclasses.run import RunStatus
from askanna.core.exceptions import GetError
from askanna.gateways.job import JobGateway

from .mixins import ListMixin

__all__ = [
    "JobSDK",
]


class JobSDK(ListMixin):
    """Management of jobs in AskAnna
    This class is a wrapper around the JobGateway and can be used to manage jobs in Python.
    """

    gateway = JobGateway()
    job_suuid = None

    def _get_job_suuid(self) -> str:
        if not self.job_suuid:
            raise ValueError("No job SUUID set")

        return self.job_suuid

    def list(
        self,
        project_suuid: Optional[str] = None,
        workspace_suuid: Optional[str] = None,
        number_of_results: int = 100,
        order_by: Optional[str] = None,
        search: Optional[str] = None,
    ) -> List[Job]:
        """List all jobs with filter and order options

        Args:
            project_suuid (str, optional): Project SUUID to filter for jobs in a project. Defaults to None.
            workspace_suuid (str, optional): Workspace SUUID to filter for jobs in a workspace. Defaults to None.
            number_of_results (int): Number of jobs to return. Defaults to 100.
            order_by (str, optional): Order by field(s).
            search (str, optional): Search for a specific job.

        Raises:
            GetError: Error based on response status code with the error message from the API

        Returns:
            List[Job]: A list of jobs. List items are of type Job dataclass.
        """
        return super().list(
            number_of_results=number_of_results,
            order_by=order_by,
            other_query_params={
                "project_suuid": project_suuid,
                "workspace_suuid": workspace_suuid,
                "search": search,
            },
        )

    def get(self, job_suuid: str) -> Job:
        """Get information of a job

        Args:
            job_suuid (str): SUUID of the job

        Raises:
            GetError: Error based on response status code with the error message from the API

        Returns:
            Job: Job information in a Job dataclass
        """
        return self.gateway.detail(job_suuid)

    def get_job_by_name(
        self,
        job_name: str,
        project_suuid: Optional[str] = None,
    ) -> Job:
        """Get information of a job by searching on the job name. This only works if the job name is unique, or unique
        within a project.

        Args:
            job_name (str): Name of the job you are looking for
            project_suuid (str, optional): Project SUUID to filter for jobs in a project. Defaults to None.

        Raises:
            GetError: If the job name is not unique or not found

        Returns:
            Job: Job information in a Job dataclass
        """
        job_list = self.list(project_suuid=project_suuid, search=job_name)

        # The above list contain all jobs where the name of the job contains the job name we are looking for.
        # We need to filter on the exact name.
        matching_jobs = list(filter(lambda x: x.name == job_name, job_list))
        if len(matching_jobs) == 0:
            raise GetError("A job with this name is not available. Did you push your code?")
        if len(matching_jobs) > 1:
            if not project_suuid:
                raise GetError(
                    "There are multiple jobs with the same name. You can narrow the selection by providing the "
                    "project SUUID."
                )
            raise GetError(
                "There are multiple jobs with the same name. This could happen if you changed names of the job "
                "manually. Please make sure the job names are unique."
            )

        return matching_jobs[0]

    def run_request(
        self,
        job_suuid: Optional[str] = None,
        data: Optional[dict] = None,
        name: Optional[str] = None,
        description: Optional[str] = None,
        project_suuid: Optional[str] = None,
        job_name: Optional[str] = None,
    ) -> RunStatus:
        """Do a request to run a job on the AskAnna platform

        Args:
            job_suuid (str, optioanl): SUUID of the job you want to start a run of. job_suuid or job_name is required.
            data (dict, optional): Data to pass to the job.
            name (str, optional): Optionally give the run a name
            description (str, optional): Optionally give the run a description
            project_suuid (str, optional): Project SUUID to filter for jobs in a project. Defaults to None.
            job_name (str, optional): Name of the job you want to start a run of. job_name or job_suuid is required.

        Raises:
            PostError: Error based on response status code with the error message from the API

        Returns:
            RunStatus: The run status information in a RunStatus dataclass
        """
        assert job_suuid or job_name, "To start a run we need at least a job SUUID or job name"
        assert not (job_suuid and job_name), "Parameters 'job_suuid' and 'job_name' are both set. Please only set one."

        if job_name:
            project_suuid = project_suuid or config.project.project_suuid
            job_suuid = self.get_job_by_name(job_name=job_name, project_suuid=project_suuid).suuid
        if not job_suuid:
            raise ValueError("No job SUUID set")

        return self.gateway.run_request(
            job_suuid=job_suuid,
            data=data,
            name=name,
            description=description,
        )

    def change(
        self,
        job_suuid: str,
        name: Optional[str] = None,
        description: Optional[str] = None,
    ) -> Job:
        """Change the name and/or description of a job

        Args:
            job_suuid (str): SUUID of the job you want to change the information of
            name (str, optional): New name for the project. Defaults to None.
            description (str, optional): New description of the project. Defaults to None.

        Raises:
            ValueError: Error when visibility is not "PUBLIC" or "PRIVATE"
            ValueError: Error if none of the arguments 'name', 'description' or 'visibility' are provided
            PatchError: Error based on response status code with the error message from the API

        Returns:
            Project: The changed project information in a Project dataclass
        """
        return self.gateway.change(
            job_suuid=job_suuid,
            name=name,
            description=description,
        )

    def delete(self, job_suuid: str) -> bool:
        """Delete a job

        Args:
            job_suuid (str): SUUID of the job you want to delete

        Raises:
            DeleteError: Error based on response status code with the error message from the API

        Returns:
            bool: True if the job was succesfully deleted
        """
        return self.gateway.delete(job_suuid=job_suuid)



import datetime
from dataclasses import dataclass
from typing import Dict, Optional

from dateutil import parser as dateutil_parser

from .base import VISIBILITY
from .relation import UserRelation


@dataclass
class Workspace:
    suuid: str
    name: str
    description: str
    visibility: VISIBILITY
    created_by: Optional[UserRelation]
    permission: dict
    is_member: bool
    created_at: datetime.datetime
    modified_at: datetime.datetime

    def __str__(self):
        return f"Workspace: {self.name} ({self.suuid})"

    @classmethod
    def from_dict(cls, data: Dict) -> "Workspace":
        data["created_at"] = dateutil_parser.parse(data["created_at"])
        data["modified_at"] = dateutil_parser.parse(data["modified_at"])

        created_by = UserRelation.from_dict(data["created_by"]) if data["created_by"] else None
        del data["created_by"]

        return cls(created_by=created_by, **data)



from dataclasses import dataclass
from typing import Dict

from .base import MembershipRole


@dataclass
class BaseRelation:
    relation: str
    suuid: str
    name: str

    @classmethod
    def from_dict(cls, data: Dict[str, str]):
        return cls(**data)


@dataclass
class WorkspaceRelation(BaseRelation):
    ...


@dataclass
class ProjectRelation(BaseRelation):
    ...


@dataclass
class PackageRelation(BaseRelation):
    ...


@dataclass
class JobRelation(BaseRelation):
    ...


@dataclass
class RunRelation(BaseRelation):
    ...
    suuid: str
    name: str


@dataclass
class UserRelation:
    relation: str
    suuid: str

    @classmethod
    def from_dict(cls, data: Dict[str, str]) -> "UserRelation":
        return cls(**data)


@dataclass
class CreatedByRelation(BaseRelation):
    job_title: str
    role: MembershipRole
    status: str

    @classmethod
    def from_dict(cls, data: Dict) -> "CreatedByRelation":
        role = MembershipRole(**data["role"])
        del data["role"]

        return cls(role=role, **data)


@dataclass
class CreatedByWithAvatarRelation(CreatedByRelation):
    avatar: dict

    @classmethod
    def from_dict(cls, data: Dict) -> "CreatedByWithAvatarRelation":
        role = MembershipRole(**data["role"])
        del data["role"]

        return cls(role=role, **data)


@dataclass
class PayloadRelation(BaseRelation):
    size: int
    lines: int



from typing import List, Optional

from askanna.core.dataclasses.base import VISIBILITY
from askanna.core.dataclasses.workspace import Workspace
from askanna.gateways.workspace import WorkspaceGateway

from .mixins import ListMixin

__all__ = [
    "WorkspaceSDK",
]


class WorkspaceSDK(ListMixin):
    """Management of workspaces in AskAnna
    This class is a wrapper around the WorkspaceGateway and can be used to manage workspaces in Python.
    """

    gateway = WorkspaceGateway()

    def list(
        self,
        number_of_results: int = 100,
        order_by: Optional[str] = None,
        search: Optional[str] = None,
        is_member: Optional[bool] = None,
        visibility: Optional[VISIBILITY] = None,
    ) -> List[Workspace]:
        """List all workspaces

        Args:
            number_of_results (int): Number of workspaces to return. Defaults to 100.
            order_by (str, optional): Order by field(s).
            search (str, optional): Search for a specific workspace.
            is_member (bool, optional): Filter on workspaces where the authenticated user is a member.
            visibility ("PRIVATE" or "PUBLIC", optional): Filter on workspaces with a specific visibility.

        Raises:
            GetError: Error based on response status code with the error message from the API

        Returns:
            List[Workspace]: A list of workspaces. List items are of type Workspace dataclass.
        """
        return super().list(
            number_of_results=number_of_results,
            order_by=order_by,
            other_query_params={
                "search": search,
                "is_member": is_member,
                "visibility": visibility,
            },
        )

    def get(self, workspace_suuid: str) -> Workspace:
        """Get information of a workspace

        Args:
            workspace_suuid (str): SUUID of the workspace

        Raises:
            GetError: Error based on response status code with the error message from the API

        Returns:
            Workspace: Workspace information in a Workspace dataclass
        """
        return self.gateway.detail(workspace_suuid)

    def create(self, name: str, description: str = "", visibility: str = "PRIVATE") -> Workspace:
        """Create a new workspace

        Args:
            name (str): Name of the new workspace
            description (str, optional): Description for the new workspace. Defaults to "" (empty string).
            visibility ("PUBLIC" or "PRIVATE", optional): Visibility of the new workspace. Defaults to "PRIVATE".

        Raises:
            ValueError: Error when visibility is not "PUBLIC" or "PRIVATE"
            CreateError: Error based on response status code with the error message from the API

        Returns:
            Workspace: The information of the newly created workspace in a Workspace dataclass
        """
        return self.gateway.create(
            name=name,
            description=description,
            visibility=visibility,
        )

    def change(
        self,
        workspace_suuid: str,
        name: Optional[str] = None,
        description: Optional[str] = None,
        visibility: Optional[str] = None,
    ) -> Workspace:
        """Change the name, description and/or visibility of a workspace

        Args:
            workspace_suuid (str): SUUID of the workspace you want to change the information of
            name (str, optional): New name for the workspace. Defaults to None.
            description (str, optional): New description of the workspace. Defaults to None.
            visibility ("PUBLIC" or "PRIVATE", optional): New visibility of the workspace. Defaults to None.

        Raises:
            ValueError: Error when visibility is not "PUBLIC" or "PRIVATE"
            ValueError: Error if none of the arguments 'name', 'description' or 'visibility' are provided
            PatchError: Error based on response status code with the error message from the API

        Returns:
            Workspace: The changed workspace information in a Workspace dataclass
        """
        return self.gateway.change(
            workspace_suuid=workspace_suuid,
            name=name,
            description=description,
            visibility=visibility,
        )

    def delete(self, workspace_suuid: str) -> bool:
        """Delete a workspace

        Args:
            workspace_suuid (str): SUUID of the workspace you want to delete

        Raises:
            DeleteError: Error based on response status code with the error message from the API

        Returns:
            bool: True if the workspace was succesfully deleted
        """
        return self.gateway.delete(workspace_suuid)



import datetime
import json
from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional, Union

from dateutil import parser as dateutil_parser

from askanna.core.exceptions import MultipleObjectsReturnedError
from askanna.core.utils.object import json_serializer

from .base import Label
from .relation import (
    CreatedByRelation,
    CreatedByWithAvatarRelation,
    JobRelation,
    PayloadRelation,
    ProjectRelation,
    RunRelation,
    WorkspaceRelation,
)

try:
    from typing import Literal
except ImportError:  # pragma: no cover
    from typing_extensions import Literal

STATUS = Literal["queued", "running", "finished", "failed"]
TRIGGER = Literal["api", "cli", "python-sdk", "webui", "schedule", "worker"]


@dataclass
class Variable:
    name: str
    value: Any
    type: str

    def __str__(self):
        return f"{self.name}: {self.value}"

    def to_dict(self) -> Dict:
        return {
            "name": self.name,
            "value": self.value,
            "type": self.type,
        }


@dataclass
class VariableObject:
    variable: Variable
    label: List[Label] = field(default_factory=list)
    run_suuid: Optional[str] = None
    created_at: Optional[datetime.datetime] = None

    def __post_init__(self):
        if not self.created_at:
            # The created time is set with timezome set to UTC so that we don't have to figure out what the local
            # timezone is. UTC is the default timezone for the API.
            self.created_at = datetime.datetime.now(datetime.timezone.utc)

    def __str__(self):
        return f"{self.variable} ({len(self.label)} label" + ("s" if len(self.label) != 1 else "") + ")"

    def __repr__(self):
        return f"VariableObject({self.variable})"

    def to_dict(self) -> Dict:
        return {
            "variable": self.variable.to_dict(),
            "label": [label.to_dict() for label in self.label],
            "run_suuid": self.run_suuid,
            "created_at": self.created_at,
        }

    @classmethod
    def from_dict(cls, data: Dict) -> "VariableObject":
        return cls(
            variable=Variable(
                name=data["variable"]["name"],
                value=data["variable"]["value"],
                type=data["variable"]["type"],
            ),
            label=[Label(**label) for label in data["label"]],
            run_suuid=data["run_suuid"],
            created_at=dateutil_parser.isoparse(data["created_at"]),
        )


@dataclass
class VariableList:
    variables: List[VariableObject] = field(default_factory=list)

    def __len__(self):
        return len(self.variables)

    def __iter__(self):
        yield from self.variables

    def __getitem__(self, row):
        return self.variables[row]

    def __str__(self):
        return f"List of {len(self.variables)} variable" + ("s" if len(self.variables) != 1 else "")

    def __repr__(self):
        return f"VariableList({len(self.variables)} variable" + ("s" if len(self.variables) != 1 else "") + ")"

    def append(self, variable: VariableObject):
        self.variables.append(variable)

    def get(self, name) -> Union[VariableObject, None]:
        variables_filtered = self.filter(name)
        if len(variables_filtered) == 1:
            return variables_filtered[0]
        if len(variables_filtered) > 1:
            raise MultipleObjectsReturnedError(
                f"Found multiple variables matching name '{name}', please use the method .filter(name=\"{name}\")."
            )
        return None

    def filter(self, name) -> List[VariableObject]:
        variables_filtered = list(filter(lambda x: x.variable.name == name, self.variables))
        return variables_filtered

    def to_dict(self) -> List[dict]:
        return [variable.to_dict() for variable in self.variables]

    def to_json(self) -> str:
        return json.dumps(self.to_dict(), default=json_serializer)


@dataclass
class Metric(Variable):
    pass


@dataclass
class MetricObject:
    metric: Metric
    label: List[Label] = field(default_factory=list)
    run_suuid: Optional[str] = None
    created_at: Optional[datetime.datetime] = None

    def __post_init__(self):
        if not self.created_at:
            # The created time is set with timezome set to UTC so that we don't have to figure out what the local
            # timezone is. UTC is the default timezone for the API.
            self.created_at = datetime.datetime.now(datetime.timezone.utc)

    def __str__(self):
        return f"{self.metric} ({len(self.label)} label" + ("s" if len(self.label) != 1 else "") + ")"

    def __repr__(self):
        return f"MetricObject({self.metric})"

    def to_dict(self) -> Dict:
        return {
            "metric": self.metric.to_dict(),
            "label": [label.to_dict() for label in self.label],
            "run_suuid": self.run_suuid,
            "created_at": self.created_at,
        }

    @classmethod
    def from_dict(cls, data: Dict) -> "MetricObject":
        return cls(
            metric=Metric(
                name=data["metric"]["name"],
                value=data["metric"]["value"],
                type=data["metric"]["type"],
            ),
            label=[Label(**label) for label in data["label"]],
            run_suuid=data["run_suuid"],
            created_at=dateutil_parser.isoparse(data["created_at"]),
        )


@dataclass
class MetricList:
    metrics: List[MetricObject] = field(default_factory=list)

    def __len__(self):
        return len(self.metrics)

    def __iter__(self):
        yield from self.metrics

    def __getitem__(self, row):
        return self.metrics[row]

    def __str__(self):
        return f"List of {len(self.metrics)} metric" + ("s" if len(self.metrics) != 1 else "")

    def __repr__(self):
        return f"MetricList({len(self.metrics)} metric" + ("s" if len(self.metrics) != 1 else "") + ")"

    def append(self, metric: MetricObject):
        self.metrics.append(metric)

    def get(self, name) -> Union[MetricObject, None]:
        metrics_filtered = self.filter(name)
        if len(metrics_filtered) == 1:
            return metrics_filtered[0]
        if len(metrics_filtered) > 1:
            raise MultipleObjectsReturnedError(
                f"Found multiple metrics matching name '{name}', please use the method .filter(name=\"{name}\")."
            )
        return None

    def filter(self, name) -> List[MetricObject]:
        metrics_filtered = list(filter(lambda x: x.metric.name == name, self.metrics))
        return metrics_filtered

    def to_dict(self) -> List[dict]:
        return [metric.to_dict() for metric in self.metrics]

    def to_json(self) -> str:
        return json.dumps(self.to_dict(), default=json_serializer)


@dataclass
class Run:
    suuid: str
    name: str
    description: str

    status: STATUS
    duration: int

    trigger: TRIGGER
    created_by: CreatedByWithAvatarRelation

    package: dict
    payload: Optional[PayloadRelation]
    result: dict
    artifact: dict
    metrics_meta: dict
    variables_meta: dict
    log: dict
    environment: dict

    job: JobRelation
    project: ProjectRelation
    workspace: WorkspaceRelation

    created_at: datetime.datetime
    modified_at: datetime.datetime
    started_at: Optional[datetime.datetime] = None
    finished_at: Optional[datetime.datetime] = None

    metrics: Optional[MetricList] = None
    variables: Optional[VariableList] = None

    def __str__(self):
        summary = {
            "suuid": self.suuid,
            "name": self.name,
            "status": self.status,
            "duration": self.duration,
            "result": self.result,
            "metrics": self.metrics_meta.get("count", 0),
            "variables": self.variables_meta.get("count", 0),
        }

        return f"Run summary: {summary}"

    def __repr__(self):
        return f"Run(suuid='{self.suuid}', status='{self.status}')"

    @classmethod
    def from_dict(cls, data: Dict) -> "Run":
        data["created_at"] = dateutil_parser.parse(data["created_at"])
        data["modified_at"] = dateutil_parser.parse(data["modified_at"])

        if data["started_at"]:
            data["started_at"] = dateutil_parser.parse(data["started_at"])
        if data["finished_at"]:
            data["finished_at"] = dateutil_parser.parse(data["finished_at"])

        created_by = CreatedByWithAvatarRelation.from_dict(data["created_by"])
        del data["created_by"]
        payload = PayloadRelation.from_dict(data["payload"]) if data["payload"] else None
        del data["payload"]
        job = JobRelation.from_dict(data["job"])
        del data["job"]
        project = ProjectRelation.from_dict(data["project"])
        del data["project"]
        workspace = WorkspaceRelation.from_dict(data["workspace"])
        del data["workspace"]

        return cls(created_by=created_by, payload=payload, job=job, project=project, workspace=workspace, **data)


@dataclass
class RunStatus:
    suuid: str
    status: STATUS
    name: str
    next_url: str
    created_by: CreatedByRelation
    job: JobRelation
    project: ProjectRelation
    workspace: WorkspaceRelation
    created_at: datetime.datetime
    modified_at: datetime.datetime
    started_at: Optional[datetime.datetime] = None
    finished_at: Optional[datetime.datetime] = None
    duration: int = 0

    def __str__(self):
        if self.name:
            return f"{self.name} ({self.suuid}): {self.status}"
        return f"Run {self.suuid}: {self.status}"

    def __repr__(self):
        return f"RunStatus(suuid='{self.suuid}', status='{self.status}')"

    @classmethod
    def from_dict(cls, data: Dict) -> "RunStatus":
        data["created_at"] = dateutil_parser.isoparse(data["created_at"])
        data["modified_at"] = dateutil_parser.isoparse(data["modified_at"])

        if "started_at" in data and data["started_at"]:
            data["started_at"] = dateutil_parser.isoparse(data["started_at"])
        if "finished_at" in data and data["finished_at"]:
            data["finished_at"] = dateutil_parser.isoparse(data["finished_at"])

        created_by = CreatedByRelation.from_dict(data["created_by"])
        del data["created_by"]
        job = JobRelation.from_dict(data["job"])
        del data["job"]
        project = ProjectRelation.from_dict(data["project"])
        del data["project"]
        workspace = WorkspaceRelation.from_dict(data["workspace"])
        del data["workspace"]

        return cls(created_by=created_by, job=job, project=project, workspace=workspace, **data)


@dataclass
class ArtifactFile:
    name: str
    type: str
    size: int
    path: str
    parent: str
    last_modified: datetime.datetime

    @classmethod
    def from_dict(cls, data: Dict) -> "ArtifactFile":
        data["last_modified"] = dateutil_parser.isoparse(data["last_modified"])
        return cls(**data)


@dataclass
class ArtifactFileList:
    files: List[ArtifactFile] = field(default_factory=list)

    def __len__(self):
        return len(self.files)

    def __iter__(self):
        yield from self.files

    def __getitem__(self, row):
        return self.files[row]

    def __str__(self):
        return f"List of {len(self.files)} item" + ("s" if len(self.files) != 1 else "")

    def __repr__(self):
        return f"ArtifactFileList({len(self.files)} item" + ("s" if len(self.files) != 1 else "") + ")"


@dataclass
class ArtifactInfo:
    suuid: str

    size: int
    count_dir: int
    count_files: int

    run: RunRelation
    job: JobRelation
    project: ProjectRelation
    workspace: WorkspaceRelation

    created_at: datetime.datetime
    modified_at: datetime.datetime

    cdn_base_url: str
    files: ArtifactFileList

    @classmethod
    def from_dict(cls, data: Dict) -> "ArtifactInfo":
        data["created_at"] = dateutil_parser.isoparse(data["created_at"])
        data["modified_at"] = dateutil_parser.isoparse(data["modified_at"])

        data["files"] = ArtifactFileList(
            [ArtifactFile.from_dict(f) for f in data["files"]],
        )

        run = RunRelation.from_dict(data["run"])
        del data["run"]
        job = JobRelation.from_dict(data["job"])
        del data["job"]
        project = ProjectRelation.from_dict(data["project"])
        del data["project"]
        workspace = WorkspaceRelation.from_dict(data["workspace"])
        del data["workspace"]

        return cls(run=run, job=job, project=project, workspace=workspace, **data)



import datetime
from dataclasses import dataclass
from typing import Dict, List, Optional

from dateutil import parser as dateutil_parser

from .relation import ProjectRelation, WorkspaceRelation


@dataclass
class Package:
    suuid: str
    workspace: WorkspaceRelation
    project: ProjectRelation
    created_by: dict
    created_at: datetime.datetime
    modified_at: datetime.datetime
    name: Optional[str]
    description: str
    filename: str
    size: int
    cdn_base_url: Optional[str] = None
    files: Optional[List] = None

    @classmethod
    def from_dict(cls, data: Dict) -> "Package":
        data["created_at"] = dateutil_parser.parse(data["created_at"])
        data["modified_at"] = dateutil_parser.parse(data["modified_at"])

        workspace = WorkspaceRelation.from_dict(data["workspace"])
        del data["workspace"]
        project = ProjectRelation.from_dict(data["project"])
        del data["project"]

        return cls(workspace=workspace, project=project, **data)



from pathlib import Path
from typing import List, Optional, Union

from askanna.core.dataclasses.package import Package
from askanna.gateways.package import PackageGateway

from .mixins import ListMixin

__all__ = [
    "PackageSDK",
]


class PackageSDK(ListMixin):
    """Management of packages in AskAnna
    This class is a wrapper around the PackageGateway and can be used to manage packages in Python.
    """

    gateway = PackageGateway()
    package_suuid = None

    def _get_package_suuid(self) -> str:
        if not self.package_suuid:
            raise ValueError("No package SUUID set")

        return self.package_suuid

    def list(
        self,
        project_suuid: Optional[str] = None,
        workspace_suuid: Optional[str] = None,
        created_by_name: Optional[str] = None,
        created_by_suuid: Optional[str] = None,
        number_of_results: int = 100,
        order_by: Optional[str] = None,
        search: Optional[str] = None,
    ) -> List[Package]:
        """List all packages with filter and order options

        Args:
            project_suuid (str, optional): Project SUUID to filter for packages in a project. Defaults to None.
            workspace_suuid (str, optional): Workspace SUUID to filter for packages in a workspace. Defaults to None.
            created_by_name (str, optional): Filter packages on a created by name. Defaults to None.
            created_by_suuid (str, optional): Filter packages on a created by SUUID. Defaults to None.
            page_size (int, optional): Number of packages to return per page. Defaults to the default value of
                the backend.
            number_of_results (int): Number of packages to return. Defaults to 100.
            order_by (str, optional): Order by field(s).
            search (str, optional): Search for a specific package.

        Raises:
            GetError: Error based on response status code with the error message from the API

        Returns:
            List[Package]: A list of packages. List items are of type Package dataclass.
        """
        return super().list(
            number_of_results=number_of_results,
            order_by=order_by,
            other_query_params={
                "project_suuid": project_suuid,
                "workspace_suuid": workspace_suuid,
                "created_by_name": created_by_name,
                "created_by_suuid": created_by_suuid,
                "search": search,
            },
        )

    def info(self, package_suuid: Optional[str] = None) -> Package:
        """Get information of a package

        Args:
            package_suuid (str): SUUID of the package

        Raises:
            GetError: Error based on response status code with the error message from the API

        Returns:
            Package: Package information in a Package dataclass
        """
        if package_suuid:
            self.package_suuid = package_suuid

        return self.gateway.detail(self._get_package_suuid())

    def get(self, package_suuid: Optional[str] = None) -> Union[bytes, None]:
        """Get the content of a package

        Args:
            package_suuid (str): SUUID of the package

        Raises:
            GetError: Error based on response status code with the error message from the API

        Returns:
            bytes or None: The package in bytes
        """
        if package_suuid:
            self.package_suuid = package_suuid
        return self.gateway.download(self._get_package_suuid())

    def download(
        self,
        output_path: Union[Path, str],
        package_suuid: Optional[str] = None,
    ) -> None:
        """Download the content of a package and save it to output_path

        Args:
            output_path (Path | str): Path to download the package to
            package_suuid (str): SUUID of the package

        Raises:
            GetError: Error based on response status code with the error message from the API

        Returns:
            bytes or None: The package in bytes, or None if output_path is set
        """
        if package_suuid:
            self.package_suuid = package_suuid

        self.gateway.download(self._get_package_suuid(), output_path)



import sys
from pathlib import Path

import click

from askanna.gateways.run import RunGateway

HELP = """Get the payload for the run from AskAnna and it to a file."""
SHORT_HELP = "Get payload from AskAnna"


@click.command(help=HELP, short_help=SHORT_HELP)
@click.option(
    "--run",
    "run_suuid",
    required=True,
    envvar="AA_RUN_SUUID",
    help="The run SUUID",
)
@click.option(
    "--payload",
    "payload_suuid",
    required=True,
    envvar="AA_PAYLOAD_SUUID",
    help="The payload SUUID",
)
@click.option(
    "--output",
    "output_path",
    envvar="AA_PAYLOAD_PATH",
    default="/input/payload.json",
    show_default=True,
    type=click.Path(path_type=Path),
)
def cli(run_suuid, payload_suuid, output_path):
    try:
        RunGateway().payload(run_suuid, payload_suuid, output_path)
    except Exception as e:
        click.echo(f"Something went wrong getting the payload. The error message received:\n  {e}", err=True)
        sys.exit(1)



import os
import sys

import click
import yaml

from askanna.cli.utils import ask_which_workspace
from askanna.config import config
from askanna.core.dataclasses.project import Project
from askanna.sdk.project import ProjectSDK

HELP = """
This command will allow you to create an AskAnna project in your current directory
"""

SHORT_HELP = "Create a project in the current directory"


class CreateProject:
    def __init__(self, name: str = ""):
        self.name = name

    def cli(self, workspace_suuid: str = "", description: str = "") -> Project:
        if not self.name:
            click.echo(
                "Hi! It is time to create a new project in AskAnna. "
                "We start with some information about the project.\n"
            )
            self.name = click.prompt("Project name", type=str)

            if not description:
                description = click.prompt("Project description", type=str, default="", show_default=False)

        if not workspace_suuid:
            workspace = ask_which_workspace("In which workspace do you want to create the new project?")
            workspace_suuid = workspace.suuid

        try:
            project = ProjectSDK().create(workspace_suuid=workspace_suuid, name=self.name, description=description)
        except Exception as e:
            click.echo(f"Something went wrong while creating the project:\n  {e}", err=True)
            sys.exit(1)
        else:
            click.echo(f"\nYou successfully created project '{project.name}' with SUUID '{project.suuid}'")
            return project


@click.command(help=HELP, short_help=SHORT_HELP)
@click.argument("name", required=False)
@click.option(
    "--workspace",
    "-w",
    show_default=True,
    help="Workspace SUUID where you want to create the project",
)
@click.option(
    "--description",
    "-d",
    default="",
    help="Description of the project",
)
def cli(name, workspace, description):
    cwd = os.getcwd()
    askanna_project_file = os.path.join(cwd, "askanna.yml")
    if os.path.exists(askanna_project_file):
        click.echo(
            "This directory already has an 'askanna.yml' file. If you continue, we will create a new project in "
            "AskAnna."
        )
        click.echo(
            "We will not update your 'askanna.yml' file with the push-target of the new project. If you continue, you "
            "need to add the push-target yourself.\n"
        )
        click.confirm("Do you want to continue without updating the 'askanna.yml'?", abort=True)
        click.echo("")

    project_creator = CreateProject(name=name)
    project_info = project_creator.cli(workspace_suuid=workspace, description=description)

    if config.server.ui:
        project_url = config.server.ui + "/" + project_info.workspace.suuid + "/project/" + project_info.suuid + "/"
    else:
        project_url = config.server.remote + "/v1/project/" + project_info.suuid + "/"

    if not os.path.exists(askanna_project_file):
        with open(askanna_project_file, "w") as pf:
            pf.write(yaml.dump({"push-target": project_url}, indent=2))
    else:
        click.echo(
            "\nTo be able to push your code to AskAnna, you need to update your local `askanna.yml` file. "
            "In the `askanna.yml` set or update the push-target with:"
        )
        click.echo(f"push-target: {project_url}")

    if config.server.ui:
        click.echo("\nWe have setup the new project. You can open your project at:")
        click.echo(project_url)
    else:
        click.echo("\nWe have setup the new project.")
    click.echo(
        "\nAs a first step you can configure your first job for this project in the `askanna.yml` file."
        "\nOnes you are done, you can push your code to AskAnna with the command:\n"
        "\n"
        "  askanna push\n"
        "\n"
        "More info about creating jobs: https://docs.askanna.io/job/create-job/\n"
        "\n"
        "Success with your project!"
    )



# Code for exiting in iPython. Inspiration from https://stackoverflow.com/a/40222538
# exit_register runs at the end of ipython %run or the end of the Python interpreter
try:
    ip = get_ipython()  # type: ignore
except NameError:
    from atexit import register as exit_register  # noqa: F401
else:  # pragma: no cover
    from functools import wraps

    def exit_register(func, *args, **kwargs):
        """Decorator that registers a post_execute for IPython. After its execution, it unregisters itself for
        subsequent runs."""

        @wraps(func)
        def wrapper():
            func(*args, **kwargs)
            ip.events.unregister("post_execute", wrapper)

        ip.events.register("post_execute", wrapper)
        return wrapper



import os
from typing import Any, Optional, Union

import click

from askanna.core.collector import MetricCollector, VariableCollector
from askanna.core.dataclasses.base import Label
from askanna.core.dataclasses.run import Metric, MetricObject, Variable, VariableObject
from askanna.core.utils.at_exit import exit_register
from askanna.core.utils.main import make_label_list
from askanna.core.utils.object import (
    get_type,
    object_fullname,
    prepare_and_validate_value,
    value_not_empty,
)

__all__ = [
    "track_metric",
    "track_metrics",
    "track_variable",
    "track_variables",
]

# Start metric and variable collection for the current session and register save collections on exit
metric_collector = MetricCollector(run_suuid=os.getenv("AA_RUN_SUUID"))
variable_collector = VariableCollector(run_suuid=os.getenv("AA_RUN_SUUID"))


@exit_register
def at_exit_save_metrics():
    if len(metric_collector) > 0:
        metric_collector.save()
    if len(variable_collector) > 0:
        variable_collector.save()


def track_metric(name: str, value: Any = None, label: Optional[Union[str, list, dict]] = None) -> None:
    if value_not_empty(value):
        value, valid = prepare_and_validate_value(value)
        if not valid:
            click.echo(
                f"AskAnna cannot store the datatype '{object_fullname(value)}'. Metric not stored for '{name}' with "
                f"value '{value}' and labels '{label}'.",
                err=True,
            )
            return
        metric = Metric(name=name, value=value, type=get_type(value))
    else:
        metric = Metric(name=name, value=None, type="tag")

    # Add value to track queue
    metric_object = MetricObject(
        metric=metric,
        label=make_label_list(label) if label else [],
    )
    metric_collector.add(metric_object)


def track_metrics(metrics: dict, label: Optional[Union[str, list, dict]] = None) -> None:
    """
    Transform many metrics to individual metrics using track_metric
    """
    for name, value in metrics.items():
        track_metric(name, value, label)


def track_variable(name: str, value: Any = None, label: Optional[Union[str, list, dict]] = None) -> None:
    if value_not_empty(value):
        value, valid = prepare_and_validate_value(value)
        if not valid:
            click.echo(
                f"AskAnna cannot store the datatype '{object_fullname(value)}'. Variable not stored for '{name}' with "
                f"value '{value}' and labels '{label}'.",
                err=True,
            )
            return

    # We don't want to store secrets, so filter them and replace value if it is sensitive information
    variable_name = name.upper()
    is_masked = any(
        [
            "KEY" in variable_name,
            "TOKEN" in variable_name,
            "SECRET" in variable_name,
            "PASSWORD" in variable_name,
        ]
    )
    if is_masked:
        value = "***masked***"

    # Add value to track queue
    if value_not_empty(value):
        variable = Variable(name=name, value=value, type=get_type(value))
    else:
        variable = Variable(name=name, value=None, type="tag")

    labels = [Label(name="source", value="run", type="string")]
    if label:
        labels.extend(make_label_list(label))
    if is_masked:
        labels.append(Label(**{"name": "is_masked", "value": None, "type": "tag"}))

    variable = VariableObject(variable=variable, label=labels)
    variable_collector.add(variable)


def track_variables(variables: dict, label: Union[str, list, dict] = "") -> None:
    """
    Transform many variables to individual variables using track_variable
    """
    for name, value in variables.items():
        track_variable(name, value, label)



import sys

import click

from askanna.cli.utils import ask_which_project, ask_which_workspace
from askanna.config import config
from askanna.config.utils import string_format_datetime
from askanna.core.exceptions import GetError, PatchError
from askanna.sdk.project import ProjectSDK


@click.group()
def cli1():
    pass


@cli1.command(help="List projects available in AskAnna", short_help="List projects")
@click.option(
    "--workspace",
    "-w",
    "workspace_suuid",
    required=False,
    type=str,
    help="Workspace SUUID to list projects for a workspace",
)
@click.option("--search", "-s", required=False, type=str, help="Search for a specific project")
@click.option(
    "--member/--no-member",
    "-m",
    "is_member",
    default=None,
    required=False,
    type=bool,
    help="Filter on projects where the authenticated user is a member.",
)
@click.option(
    "--visibility", "-v", required=False, type=str, help="Filter on projects with visibility private or public"
)
def list(workspace_suuid, search, is_member, visibility):
    project_sdk = ProjectSDK()
    try:
        projects = project_sdk.list(
            number_of_results=100,
            workspace_suuid=workspace_suuid,
            search=search,
            is_member=is_member,
            visibility=visibility,
            order_by="workspace.name,name",
        )
    except Exception as e:
        click.echo(f"Something went wrong while listing the projects:\n  {e}", err=True)
        sys.exit(1)

    if not projects:
        click.echo("We cannot find any project.")
        sys.exit(0)

    if workspace_suuid:
        click.echo(f"The projects for workspace '{projects[0].workspace.name}' are:")
        click.echo("")
        click.echo("-------------------    -------------------------")
        click.echo("PROJECT SUUID          PROJECT NAME")
        click.echo("-------------------    -------------------------")
    else:
        click.echo("")
        click.echo("-------------------    --------------------    -------------------    -------------------------")
        click.echo("WORKSPACE SUUID        WORKSPACE NAME          PROJECT SUUID          PROJECT NAME")
        click.echo("-------------------    --------------------    -------------------    -------------------------")

    for project in projects:
        project_name = f"{project.name[:22]}..." if len(project.name) > 25 else project.name[:25]
        if workspace_suuid:
            click.echo(f"{project.suuid}    {project_name}")
        else:
            workspace_name = (
                f"{project.workspace.name[:17]}..."
                if len(project.workspace.name) > 20
                else project.workspace.name[:20]
            )
            click.echo(
                "{workspace_suuid}    {workspace_name}    {project_suuid}    {project_name}".format(
                    workspace_suuid=project.workspace.suuid,
                    workspace_name=f"{workspace_name:20}",
                    project_suuid=project.suuid,
                    project_name=project_name,
                )
            )

    if len(projects) != project_sdk.list_total_count:
        click.echo("")
        click.echo(f"Note: the first {len(projects):,} of {project_sdk.list_total_count:,} projects are shown.")

    click.echo("")


@cli1.command(help="Get information about a project", short_help="Get project info")
@click.option("--id", "-i", "project_suuid", required=False, type=str, help="Project SUUID")
def info(project_suuid):
    if not project_suuid:
        project_suuid = config.project.project_suuid

    if project_suuid:
        try:
            project = ProjectSDK().get(project_suuid=project_suuid)
        except GetError as e:
            click.echo(f"Something went wrong while getting the project:\n  {e}", err=True)
            sys.exit(1)
    else:
        workspace = ask_which_workspace(question="From which workspace do you want to get a project?")
        project = ask_which_project(question="Which project do you want to get?", workspace_suuid=workspace.suuid)

    click.echo("")
    click.echo(f"Name:            {project.name}")
    click.echo(f"SUUID:           {project.suuid}")
    click.echo(f"Description:     {project.description}")
    click.echo(f"Visibility:      {project.visibility}")
    click.echo("")
    click.echo(f"Workspace:       {project.workspace.name}")
    click.echo(f"Workspace SUUID: {project.workspace.suuid}")
    click.echo("")
    click.echo(f"Created:         {project.created_at.strftime(string_format_datetime)}")
    click.echo(f"Modified:        {project.modified_at.strftime(string_format_datetime)}")
    click.echo("")


@cli1.command(help="Change project information in AskAnna", short_help="Change project")
@click.option("--id", "-i", "project_suuid", required=False, type=str, help="Project SUUID")
@click.option("--name", "-n", required=False, type=str, help="New name to set")
@click.option("--description", "-d", required=False, type=str, help="New description to set")
@click.option("--visibility", "-v", required=False, type=str, help="New visibility to set (PRIVATE or PUBLIC)")
def change(project_suuid, name, description, visibility):
    if not project_suuid:
        project_suuid = config.project.project_suuid

        if not project_suuid:
            workspace = ask_which_workspace(question="From which workspace do you want to change a project?")
            project = ask_which_project(
                question="Which project do you want to change?", workspace_suuid=workspace.suuid
            )
            project_suuid = project.suuid

    if not name and not description and not visibility:
        if click.confirm("\nDo you want to change the name of the project?"):
            name = click.prompt("New name of the project", type=str)
        if click.confirm("\nDo you want to change the description of the project?"):
            description = click.prompt("New description of the project", type=str)
        if click.confirm("\nDo you want to change the visibility of the project?"):
            visibility = click.prompt("Visibility of the project (PRIVATE or PUBLIC)", type=str)

        click.confirm("\nDo you want to change the project?", abort=True)

    try:
        project = ProjectSDK().change(
            project_suuid=project_suuid, name=name, description=description, visibility=visibility
        )
    except PatchError as e:
        if str(e).startswith("404"):
            click.echo(f"The project SUUID '{project_suuid}' was not found", err=True)
            sys.exit(1)
        else:
            click.echo(f"Something went wrong while changing the project SUUID '{project_suuid}':\n  {e}", err=True)
            sys.exit(1)
    else:
        click.echo(f"\nYou succesfully changed project '{project.name}' with SUUID '{project.suuid}'")


@cli1.command(help="Create a new project in AskAnna", short_help="Create project")
@click.option(
    "--workspace",
    "-w",
    "workspace_suuid",
    type=str,
    help="Workspace SUUID where you want to create the project",
)
@click.option("--name", "-n", required=False, type=str, help="Name of the project")
@click.option("--description", "-d", required=False, type=str, help="Description of the project")
@click.option(
    "--visibility",
    "-v",
    required=False,
    type=str,
    default="PRIVATE",
    help="Project visibility (PRIVATE (default) or PUBLIC)",
)
def create(workspace_suuid, name, description, visibility):
    if not workspace_suuid:
        workspace = ask_which_workspace("In which workspace do you want to create the new project?")
        workspace_suuid = workspace.suuid

    if not name:
        name = click.prompt("\nName of the project", type=str)

        if not description:
            description = click.prompt("Optional description of the project", type=str, default="", show_default=False)

    click.confirm(f'\nDo you want to create the project "{name}"?', abort=True)

    try:
        project = ProjectSDK().create(
            workspace_suuid=workspace_suuid, name=name, description=description, visibility=visibility
        )
    except Exception as e:
        click.echo(f"Something went wrong while creating the project:\n  {e}", err=True)
        sys.exit(1)
    else:
        click.echo(f"\nYou successfully created project '{project.name}' with SUUID '{project.suuid}'")


@cli1.command(help="Remove a project in AskAnna", short_help="Remove project")
@click.option("--id", "-i", "project_suuid", type=str, required=True, help="Project SUUID", prompt="Project SUUID")
@click.option("--force", "-f", is_flag=True, help="Force")
def remove(project_suuid, force):
    try:
        project = ProjectSDK().get(project_suuid=project_suuid)
    except GetError as e:
        if str(e).startswith("404"):
            click.echo(f"The project SUUID '{project_suuid}' was not found", err=True)
            sys.exit(1)
        else:
            click.echo(f"Something went wrong while removing the project SUUID '{project_suuid}':\n  {e}", err=True)
            sys.exit(1)

    if not force:
        if not click.confirm(f"Are you sure to remove project SUUID '{project_suuid}' with name '{project.name}'?"):
            click.echo("Understood. We are not removing the project.")
            sys.exit(0)

    try:
        _ = ProjectSDK().delete(project_suuid=project_suuid)
    except Exception as e:
        click.echo(f"Something went wrong while removing the project SUUID '{project_suuid}':\n  {e}", err=True)
        sys.exit(1)
    else:
        click.echo(f"You removed project SUUID '{project_suuid}'")


cli = click.CommandCollection(
    sources=[cli1],
    help="Manage your projects in AskAnna",
    short_help="Manage projects in AskAnna",
)



try:
    import zoneinfo
except ImportError:  # pragma: no cover
    from backports import zoneinfo
from typing import Dict, List, Union

import click
import croniter
from email_validator import EmailNotValidError, validate_email
from tzlocal import get_localzone


def flatten(t):
    return [item for sublist in t for item in sublist]


def validate_cron_line(cron_line: str) -> bool:
    """
    We validate the cron expression with croniter
    """
    return croniter.croniter.is_valid(cron_line)


def parse_cron_line(cron_line: str) -> Union[str, None]:
    """
    parse incoming cron definition
    if it is valid, then return the cron_line, otherwise a None
    """

    if isinstance(cron_line, str):
        # we deal with cron strings
        # first check whether we need to make a "translation" for @strings

        alias_mapping = {
            "@midnight": "0 0 * * *",
            "@yearly": "0 0 1 1 *",
            "@annually": "0 0 1 1 *",
            "@monthly": "0 0 1 * *",
            "@weekly": "0 0 * * 0",
            "@daily": "0 0 * * *",
            "@hourly": "0 * * * *",
        }
        cron_line = alias_mapping.get(cron_line.strip(), cron_line.strip())

    elif isinstance(cron_line, dict):
        # we deal with dictionary
        # first check whether we have valid keys, if one invalid key is found, return None
        valid_keys = {"minute", "hour", "day", "month", "weekday"}
        invalid_keys = set(cron_line.keys()) - valid_keys
        if len(invalid_keys):
            return None
        cron_line = "{minute} {hour} {day} {month} {weekday}".format(
            minute=cron_line.get("minute", "0"),
            hour=cron_line.get("hour", "0"),
            day=cron_line.get("day", "*"),
            month=cron_line.get("month", "*"),
            weekday=cron_line.get("weekday", "*"),
        )

    if not validate_cron_line(cron_line):
        return None

    return cron_line


def parse_cron_schedule(schedule: List):
    """
    Determine and validate which format the cron defition it has, it can be one of the following:
    - * * * * *  (m, h, d, m, weekday)
    - @annotation (@yearly, @annually, @monthly, @weekly, @daily, @hourly)
    - dict (k,v with: minute,hour,day,month,weekday)
    """

    for cron_line in schedule:
        yield cron_line, parse_cron_line(cron_line)


def validate_yml_job_names(config):
    """
    Within AskAnna, we have several variables reserved and cannot be used for jobnames
    """
    reserved_keys = (
        "askanna",
        "cluster",
        "environment",
        "image",
        "job",
        "notifications",
        "project",
        "push-target",
        "timezone",
        "variables",
        "worker",
    )

    overlapping_with_reserved_keys = list(set(config.keys()).intersection(set(reserved_keys)))
    for overlap_key in overlapping_with_reserved_keys:
        is_dict = isinstance(config.get(overlap_key), dict)
        is_job = is_dict and config.get(overlap_key).get("job")
        if is_dict and is_job:
            # we do find a job definition under the name, so probably a jobdef
            click.echo(
                f"The name `{overlap_key}` cannot be used for a job.\n"
                "This name is used to configure something else in AskAnna.\n"
                f"Invalid job name: {overlap_key}",
                err=True,
            )
            return False
    return True


def validate_yml_environments(config: Dict, jobname=None) -> bool:
    """
    Validate the environment definitions (if defined)
    """
    # Do we have a global `environment` defined?
    environment = config.get("environment")
    if environment:
        if not isinstance(environment, dict):
            if jobname:
                click.echo(
                    f"Invalid definition of `environment` found in job `{jobname}`:\n"
                    f"environment: {environment}\n"
                    "\n"
                    "For environment documentation: https://docs.askanna.io/environment/",
                    err=True,
                )
            else:
                click.echo(
                    "Invalid definition of `environment` found:\n"
                    f"environment: {environment}\n"
                    "\n"
                    "For environment documentation: https://docs.askanna.io/environment/",
                    err=True,
                )
            return False
        else:
            # make sure we have at least the `image` defined
            image = environment.get("image")
            if not image:
                click.echo(
                    "`image` was not defined in the `environment`:\n"
                    f"environment: {environment}\n"
                    "\n"
                    "For environment documentation: https://docs.askanna.io/environment/",
                    err=True,
                )
                return False

    return True


def validate_yml_notifications(config: Dict, jobname=None) -> bool:
    """
    Validate the definition of `notifications` in global and job
    """
    notifications = config.get("notifications")
    if notifications and not isinstance(notifications, dict):
        if jobname:
            click.echo(
                f"Invalid definition of `notifications` found in job `{jobname}`:\n"
                f"notifications: {notifications}"
                "\n"
                "For notifications documentation: https://docs.askanna.io/job/notifications/",
                err=True,
            )
        else:
            click.echo(
                "Invalid `notifications` setting found:"
                f"notifications: {notifications}"
                "\n"
                "For notifications documentation: https://docs.askanna.io/job/notifications/",
                err=True,
            )
        return False
    elif notifications and isinstance(notifications, dict):
        # check for values of `all` and `error`, both should be dicts
        noti_all = notifications.get("all")
        noti_error = notifications.get("error")
        if any(
            [
                noti_all and not isinstance(noti_all, dict),
                noti_error and not isinstance(noti_error, dict),
            ]
        ):
            if jobname:
                click.echo(
                    f"Invalid definition of `notifications` found in job `{jobname}`:\n"
                    f"notifications: {notifications}"
                    "\n"
                    "For notifications documentation: https://docs.askanna.io/job/notifications/",
                    err=True,
                )
            else:
                click.echo(
                    "Invalid `notifications` setting found:\n"
                    f"notifications: {notifications}"
                    "\n"
                    "For notifications documentation: https://docs.askanna.io/job/notifications/",
                    err=True,
                )
            return False

        # proceed with validation of the content
        # assume we can get values out of all/email and error/email
        noti_all_values = notifications.get("all", {}).get("email", [])
        noti_error_values = notifications.get("error", {}).get("email", [])

        all_email = sorted(
            flatten(
                list(
                    map(
                        lambda x: x.split(","),
                        noti_all_values + noti_error_values,
                    )
                )
            )
        )
        for value in all_email:
            if value.startswith("${") and value.endswith("}"):
                # we found a variable not subsituted yet
                continue
            if value in ["workspace admins", "workspace members"]:
                # valid addresees
                continue

            # validate on actual content of the value whether it is an e-mail address or not
            try:
                validate_email(value)
            except EmailNotValidError:
                click.echo(
                    "Invalid `notifications/email` found:\n" f"{value}" "\n",
                    err=True,
                )
                return False

    return True


def validate_askanna_yml(config):
    """
    Given an dictionary of askanna.yml validate each component
    """
    jobs = config.items()

    global_timezone = config.get("timezone")
    # validate the global timezone
    if global_timezone:
        if global_timezone not in zoneinfo.available_timezones():
            click.echo(
                "Invalid timezone setting found in askanna.yml:\n" + f"timezone: {global_timezone}",
                err=True,
            )
            return False
        timezone_checked = True  # used later, so we only print a warning message about the timezone once
    else:
        timezone_checked = False  # used later, so we only print a warning message about the timezone once

    # validate whether the global environment definitions are correct
    if not validate_yml_environments(config):
        return False

    # validate for global notification settings
    if not validate_yml_notifications(config):
        return False

    # validate jobs names
    if not validate_yml_job_names(config):
        return False

    for jobname, job in jobs:
        if isinstance(job, dict):
            if not validate_yml_environments(job, jobname=jobname):
                return False
            if not validate_yml_notifications(job, jobname=jobname):
                return False

            schedule = job.get("schedule")
            timezone = job.get("timezone")
            if not schedule:
                continue
            # validate the schedule
            for cron_line, parsed in parse_cron_schedule(schedule):
                if not parsed:
                    click.echo(
                        f"Invalid schedule definition found in job `{job}`: {cron_line}",
                        err=True,
                    )
                    return False
            # validate the timezone if set
            timezone = job.get("timezone")
            if timezone and timezone not in zoneinfo.available_timezones():
                click.echo(
                    f"Invalid timezone setting found in job `{jobname}`:\n" + f"timezone: {timezone}",
                    err=True,
                )
                return False

            # validate the schedule
            schedule = job.get("schedule")
            if schedule:
                for cron_line, parsed in parse_cron_schedule(schedule):
                    if not parsed:
                        click.echo(
                            f"Invalid schedule definition `{cron_line}` found in job `{jobname}`",
                            err=True,
                        )
                        return False

                if not timezone and not global_timezone and not timezone_checked:
                    timezone_local = get_localzone()
                    if timezone_local != "UTC":
                        click.echo(  # noqa
                            f"""
By default, the AskAnna platform uses time zone UTC. Your current time zone is {timezone_local}.
To use your local time zone for scheduling jobs in this project, add the next line to your config in `askanna.yml`:

timezone: {timezone_local}

For more information, read the documentation: https://docs.askanna.io/job/create-job/#time-zone
"""
                        )
                    timezone_checked = True
    return True



import importlib

import click

from askanna import ASKANNA_VERSION

HELP = """
The AskAnna CLI helps you running data science projects on AskAnna.
"""

SHORT_HELP = "AskAnna CLI"

EPILOG = """
For usage and help on a specific command, run it with a --help flag, e.g.:

    askanna [COMMAMD] --help
"""


@click.group(
    help=HELP,
    short_help=SHORT_HELP,
    epilog=EPILOG,
)
@click.version_option(version=ASKANNA_VERSION, prog_name="AskAnna CLI")
def cli():
    """
    Initialize the AskAnna CLI commands
    """


commands = [
    "artifact",
    "create",
    "init",
    "job",
    "login",
    "logout",
    "project",
    "push",
    "result",
    "run",
    "variable",
    "workspace",
]

for command in commands:
    module_path = "askanna.cli." + command
    command_module = importlib.import_module(module_path)
    command_name = command.replace("_", "-")
    cli.add_command(command_module.cli, command_name)



import uuid as _uuid
from typing import Optional

# Inspiration for create suuid from uuid comes from https://pythonhosted.org/shorten/user/examples.html


def bx_encode(n: int, alphabet: str) -> str:
    """Encodes an integer 'n' in base 'len(alphabet)' with digits in 'alphabet'.

    Args:
        n (int): the integer to encode
        alphabet (str): the alphabet to use for encoding

    Raises:
        TypeError: if arguments are not of the correct type

    Returns:
        str: the encoded string
    """

    if not isinstance(n, int):
        raise TypeError("arg 'n' must be an integer")
    if not isinstance(alphabet, str):
        raise TypeError("arg 'alphabet' must be a string")

    base = len(alphabet)

    if n == 0:
        return alphabet[0]

    digits = []

    while n > 0:
        digits.append(alphabet[n % base])
        n = n // base

    digits.reverse()
    return "".join(digits)


def str_to_groups(string: str, group_size: int, n_groups: int) -> str:
    """Make a 'grouped string' from a string. The string is split into 'n_groups' groups of size 'group_size' and
    joined with a '-'.

    Args:
        string (str): the string to convert to suuid
        group_size (int): the size of each group
        n_groups (int): the number of groups

    Raises:
        ValueError: if the length of the string is not at least a multiple of 'group_size' and 'n_groups'

    Returns:
        str: a string in the suuid form
    """
    if len(string) < group_size * n_groups:
        raise ValueError("String is too short given the group size and number of groups")

    string_groups = [string[i : i + group_size] for i in range(0, len(string), group_size)]

    return "-".join(string_groups[:n_groups])


def create_suuid(uuid: Optional[_uuid.UUID] = None) -> str:
    """create_suuid will produce 16 character alphabetic tokens, split into 4 character groups. The format is:

        xxxx-xxxx-xxxx-xxxx

    When you provide a uuid, the suuid will be based on that uuid. If you don't provide a uuid, a new uuid will be
    generated.

    Args:
        uuid (uuid.UUID, optional): an uuid4 object

    Returns:
        str: a string in the suuid form
    """
    alphabet = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz"

    token_length = 16
    group_size = 4

    n_groups = int(token_length / group_size)

    # Generate a random UUID if not given
    if not uuid:
        uuid = _uuid.uuid4()

    # Convert uuid to digits with the given alphabet
    token = bx_encode(int(uuid.hex, 16), alphabet)
    # Padding with the first symbol from 'alphabet' as needed to get the desired length
    token = token.rjust(token_length, alphabet[0])

    return str_to_groups(token, group_size, n_groups)



from typing import List, Optional

from askanna.core.dataclasses.base import VISIBILITY
from askanna.core.dataclasses.project import Project
from askanna.core.exceptions import CreateError, DeleteError, GetError, PatchError
from askanna.gateways.api_client import client

from .utils import ListResponse


class ProjectListResponse(ListResponse):
    def __init__(self, data: dict):
        super().__init__(data)
        self.results: List[Project] = [Project.from_dict(project) for project in data["results"]]

    @property
    def projects(self):
        return self.results


class ProjectGateway:
    """Management of projects in AskAnna
    This is the class which act as the gateway to the API of AskAnna
    """

    def list(
        self,
        workspace_suuid: Optional[str] = None,
        page_size: Optional[int] = None,
        cursor: Optional[str] = None,
        order_by: Optional[str] = None,
        search: Optional[str] = None,
        is_member: Optional[bool] = None,
        visibility: Optional[VISIBILITY] = None,
    ) -> ProjectListResponse:
        """List all projects with filter and order options

        Args:
            workspace_suuid (str, optional): Workspace SUUID to filter for projects in a workspace. Defaults to None.
            page_size (int, optional): Number of results per page. Defaults to the default value of the backend.
            cursor (str, optional): Cursor to start the page from. Defaults to None.
            order_by (str, optional): Order by field(s). Defaults to the default value of the backend.
            search (str, optional): Search for a specific project.
            is_member (bool, optional): Filter on projects where the authenticated user is a member.
            visibility ("PRIVATE" or "PUBLIC", optional): Filter on projects with a specific visibility.

        Raises:
            GetError: Error based on response status code with the error message from the API

        Returns:
            ProjectListResponse: The response from the API with a list of projects and pagination information.
        """
        assert page_size is None or page_size > 0, "page_size must be a positive integer"
        assert is_member is None or isinstance(is_member, bool), "is_member must be a boolean"
        if visibility is not None:
            visibility = visibility.lower()  # type: ignore
            assert visibility in ["public", "private"], "visibility must be 'public' or 'private'"

        response = client.get(
            url=client.askanna_url.project.project_list(),
            params={
                "workspace_suuid": workspace_suuid,
                "page_size": page_size,
                "cursor": cursor,
                "order_by": order_by,
                "search": search,
                "is_member": is_member,
                "visibility": visibility,
            },
        )

        if response.status_code != 200:
            error_message = f"{response.status_code} - Something went wrong while retrieving the project list"
            try:
                error_message += f":\n  {response.json()}"
            except ValueError:
                pass
            raise GetError(error_message)

        return ProjectListResponse(response.json())

    def detail(self, project_suuid: str) -> Project:
        """Get information of a project

        Args:
            project_suuid (str): SUUID of the project

        Raises:
            GetError: Error based on response status code with the error message from the API

        Returns:
            Project: Project information in a Project dataclass
        """
        response = client.get(
            url=client.askanna_url.project.project_detail(project_suuid=project_suuid),
        )

        if response.status_code == 404:
            raise GetError(f"404 - The project SUUID '{project_suuid}' was not found")
        if response.status_code != 200:
            raise GetError(
                f"{response.status_code} - Something went wrong while retrieving project SUUID '{project_suuid}': "
                f"{response.json()}"
            )

        return Project.from_dict(response.json())

    def create(self, workspace_suuid: str, name: str, description: str = "", visibility: str = "PRIVATE") -> Project:
        """Create a new project

        Args:
            workspace_suuid (str): SUUID of the workspace to create the project in
            name (str): Name of the new project
            description (str, optional): Description for the new project. Defaults to "" (empty string).
            visibility (str, optional): Visibility of the new workspace. Defaults to "PRIVATE".

        Raises:
            ValueError: Error when visibility is not "PUBLIC" or "PRIVATE"
            CreateError: Error based on response status code with the error message from the API

        Returns:
            Project: The information of the newly created project in a Project dataclass
        """
        if visibility and visibility not in ["PUBLIC", "PRIVATE"]:
            raise ValueError("Visibility must be either PUBLIC or PRIVATE")

        if description is None:
            description = ""

        response = client.create(
            url=client.askanna_url.project.project(),
            json={
                "workspace_suuid": workspace_suuid,
                "name": name,
                "description": description,
                "visibility": visibility,
            },
        )

        if response.status_code != 201:
            raise CreateError(
                f"{response.status_code} - Something went wrong while creating the project: {response.json()}"
            )

        return Project.from_dict(response.json())

    def change(
        self,
        project_suuid: str,
        name: Optional[str] = None,
        description: Optional[str] = None,
        visibility: Optional[str] = None,
    ) -> Project:
        """Change the name, description and/or visibility of a project

        Args:
            project_suuid (str): SUUID of the project to change
            name (str, optional): New name for the project. Defaults to None.
            description (str, optional): New description of the project. Defaults to None.
            visibility ("PUBLIC" or "PRIVATE", optional): New visibility of the project. Defaults to None.

        Raises:
            ValueError: Error when visibility is not "PUBLIC" or "PRIVATE"
            ValueError: Error if none of the arguments 'name', 'description' or 'visibility' are provided
            PatchError: Error based on response status code with the error message from the API

        Returns:
            Project: The changed project information in a Project dataclass
        """

        changes = {}
        if name:
            changes.update({"name": name})
        if description:
            changes.update({"description": description})
        if visibility:
            if visibility not in ["PUBLIC", "PRIVATE"]:
                raise ValueError("Visibility must be either PUBLIC or PRIVATE")
            changes.update({"visibility": visibility})

        if not changes:
            raise ValueError("At least one of the parameters 'name', 'description' or 'visibility' must be set.")

        response = client.patch(
            url=client.askanna_url.project.project_detail(project_suuid=project_suuid),
            json=changes,
        )

        if response.status_code == 404:
            raise PatchError(f"404 - The project SUUID '{project_suuid}' was not found")
        if response.status_code != 200:
            raise PatchError(
                f"{response.status_code} - Something went wrong while updating the project SUUID '{project_suuid}': "
                f"{response.json()}"
            )

        return Project.from_dict(response.json())

    def delete(self, project_suuid: str) -> bool:
        """Delete a project

        Args:
            project_suuid (str): SUUID of the project you want to delete

        Raises:
            DeleteError: Error based on response status code with the error message from the API

        Returns:
            bool: True if the project was succesfully deleted
        """
        response = client.delete(
            url=client.askanna_url.project.project_detail(project_suuid),
        )

        if response.status_code == 404:
            raise DeleteError(f"404 - The project SUUID '{project_suuid}' was not found")
        if response.status_code != 204:
            raise DeleteError(
                f"{response.status_code} - Something went wrong while deleting the project SUUID '{project_suuid}': "
                f"{response.json()}"
            )

        return True



import datetime
from typing import Any, Tuple

supported_data_types = {
    # primitive types
    "bool": "boolean",
    "int": "integer",
    "float": "float",
    "str": "string",
    # complex types
    "datetime.datetime": "datetime",
    "datetime.time": "time",
    "datetime.date": "date",
    "dict": "dictionary",
    "list": "list",
    "tag": "tag",
}

# If numpy as installed, we add the numpy types to the supported types
try:
    import numpy as np  # noqa: F401
except ImportError:  # pragma: no cover
    NUMPY_INSTALLED = False
else:
    NUMPY_INSTALLED = True
    supported_data_types.update(
        {
            # Source Numpy documentation, a.o. https://numpy.org/doc/stable/user/basics.types.html
            "numpy.bool_": "boolean",
            "numpy.intc": "integer",
            "numpy.int": "integer",
            "numpy.int_": "integer",
            "numpy.uint": "integer",
            "numpy.short": "integer",
            "numpy.ushort": "integer",
            "numpy.longlong": "integer",
            "numpy.ulonglong": "integer",
            "numpy.half": "float",
            "numpy.float16": "float",
            "numpy.single": "float",
            "numpy.double": "float",
            "numpy.longdouble": "float",
            "numpy.csingle": "float",
            "numpy.cdouble": "float",
            "numpy.clongdouble": "float",
            "numpy.str": "string",
            "numpy.str_": "string",
            # List type
            "numpy.array": "list",
            "numpy.ndarray": "list",
            # C-derived names and more "complex" types
            # Source: https://numpy.org/doc/stable/reference/arrays.scalars.html#sized-aliases
            "numpy.bool8": "boolean",
            "numpy.int8": "integer",
            "numpy.int16": "integer",
            "numpy.int32": "integer",
            "numpy.int64": "integer",
            "numpy.uint8": "integer",
            "numpy.uint16": "integer",
            "numpy.uint32": "integer",
            "numpy.uint64": "integer",
            "numpy.intp": "integer",
            "numpy.uintp": "integer",
            "numpy.float32": "float",
            "numpy.float64": "float",
            "numpy.float_": "float",
            "numpy.complex64": "float",
            "numpy.complex128": "float",
            "numpy.complex192": "float",
            "numpy.complex256": "float",
            "numpy.complex_": "float",
            "numpy.longfloat": "float",
            "numpy.singlecomplex": "float",
            "numpy.cfloat": "float",
            "numpy.longcomplex": "float",
            "numpy.clongfloat": "float",
            "numpy.unicode_": "string",
        }
    )


def object_fullname(o):
    # https://stackoverflow.com/a/2020083
    # o.__module__ + "." + o.__class__.__qualname__ is an example in
    # this context of H.L. Mencken's "neat, plausible, and wrong."
    # Python makes no guarantees as to whether the __module__ special
    # attribute is defined, so we take a more circumspect approach.
    # Alas, the module name is explicitly excluded from __qualname__
    # in Python 3.

    module = o.__class__.__module__
    if module is None or module == str.__class__.__module__:
        return o.__class__.__name__  # Avoid reporting __builtin__
    else:
        return module + "." + o.__class__.__name__


def get_type(value: Any) -> str:
    """
    Return the full name of the type, if not listed, return the typename from the input
    """
    typename = object_fullname(value)
    dtype = supported_data_types.get(typename, typename)

    if typename in ("numpy.array", "numpy.ndarray"):
        dtype_list = supported_data_types.get("numpy." + value.dtype.name)
        if dtype_list:
            dtype = "list_" + dtype_list
    elif typename == "list":
        dtype_list = None
        for val in value:
            if type(val) == bool and dtype_list in (None, "boolean"):
                dtype_list = "boolean"
            elif type(val) == int and dtype_list in (None, "integer"):
                dtype_list = "integer"
            elif type(val) in (int, float) and dtype_list in (None, "integer", "float"):
                dtype_list = "float"
            elif type(val) == str and dtype_list in (None, "string"):
                dtype_list = "string"
            elif type(val) == datetime.datetime and dtype_list in (None, "datetime"):
                dtype_list = "datetime"
            elif type(val) == datetime.time and dtype_list in (None, "time"):
                dtype_list = "time"
            elif type(val) == datetime.date and dtype_list in (None, "date"):
                dtype_list = "date"
            else:
                dtype_list = "mixed"
                break

        if dtype_list and dtype_list != "mixed":
            dtype = "list_" + dtype_list

    return dtype


def validate_value(value: Any) -> bool:
    """
    Validate whether the value set is supported
    """

    return object_fullname(value) in supported_data_types


def transform_value(value: Any) -> Tuple[Any, bool]:
    """
    Transform values in support datatypes
    """
    if object_fullname(value) == "range":
        return list(value), True

    return value, False


def prepare_and_validate_value(value: Any) -> Tuple[Any, bool]:
    """
    Validate value and if necessary transform values in support datatypes
    """
    if validate_value(value):
        return value, True

    # Try to transform the value
    value, transform = transform_value(value)
    if transform:
        return value, True

    return value, False


def value_not_empty(value: Any) -> bool:
    """
    Check if the value is not empty
    """
    if value.__class__.__module__ == "numpy":
        try:
            if value.any() or isinstance(value, np.bool_):
                return True
        except np.core._exceptions.UFuncTypeError:  # type: ignore
            pass
    if value or isinstance(value, bool):
        return True
    return False


def serialize_numpy_for_json(obj):
    # the o.item() is a generic method on each numpy dtype.
    if isinstance(obj, np.generic):
        return obj.item()
    if isinstance(obj, np.ndarray):
        return obj.tolist()
    return obj


def json_serializer(obj):
    if isinstance(obj, (datetime.time, datetime.date, datetime.datetime)):
        return obj.isoformat()
    if NUMPY_INSTALLED:
        return serialize_numpy_for_json(obj)
    return obj



import datetime
from dataclasses import dataclass
from typing import Any, Dict

try:
    from typing import Literal
except ImportError:  # pragma: no cover
    from typing_extensions import Literal

VISIBILITY = Literal["private", "public", "PRIVATE", "PUBLIC"]


@dataclass
class Label:
    name: str
    value: Any
    type: str

    def __str__(self):
        return f"{self.name}: {self.value} [type: {self.type}]"

    def to_dict(self, yes=True) -> Dict:
        return {"name": self.name, "value": self.value, "type": self.type}


@dataclass
class User:
    suuid: str
    name: str
    email: str
    is_active: bool
    date_joined: datetime.datetime
    last_login: datetime.datetime


@dataclass
class MembershipRole:
    name: str
    code: str



import json
import tempfile
import uuid
from pathlib import Path
from typing import Optional, Union

import click

from askanna.core.dataclasses.run import (
    MetricList,
    MetricObject,
    VariableList,
    VariableObject,
)
from askanna.core.utils.suuid import create_suuid
from askanna.gateways.run import RunGateway


class CollectorTemplate:
    """
    Base class for collectors
    """

    # To use the CollectorTemplate, you need to create a class that inherits from it. This class needs to implement the
    # following attributes:
    # - collector_type
    # - file_name

    # property:
    # - data_collection

    # And methods:
    # - append_from_dict (with one input attribute of type dict)
    # - save_to_askanna

    collector_type = "template"
    file_name = "collector.json"

    def __init__(self, run_suuid: Optional[str] = None):
        if self.collector_type == "template":
            assert self.collector_type != "template", (
                "CollectorTemplate is an abstract base class. You should subclass it and set 'collector_type' to "
                "a value that is not 'template'."
            )

        self.run_suuid = run_suuid
        if not self.run_suuid:
            self.local_suuid = create_suuid(uuid.uuid4())

        self.collector_file = Path(tempfile.gettempdir(), "askanna/run", self.suuid, self.file_name)

        if self.run_suuid:
            self._restore_session()

        self.changed = False

    @property
    def data_collection(self):
        raise NotImplementedError(f"Please implement 'data_collection' for  {self.__class__.__name__}")

    def append_from_dict(self, *args) -> None:
        raise NotImplementedError(f"Please implement 'append_from_dict' for  {self.__class__.__name__}")

    def save_to_askanna(self):
        raise NotImplementedError(f"Please implement 'save_to_askanna' for  {self.__class__.__name__}")

    @property
    def suuid(self) -> str:
        """
        Return the run_suuid if exist, otherwise the local_suuid
        """
        return self.run_suuid or self.local_suuid

    def __len__(self) -> int:
        return len(self.data_collection)

    def add(self, object: Union[MetricObject, VariableObject]) -> None:
        if not object.run_suuid:
            object.run_suuid = self.suuid
        elif object.run_suuid != self.suuid:
            raise ValueError(
                f"The run SUUID '{object.run_suuid}' of the {self.collector_type} does not match the run SUUID "
                f"'{self.suuid}' of the session where the collector is initialized."
            )

        self.data_collection.append(object)
        self.save_local()
        self.changed = True

    def _restore_session(self):
        """
        Restore the run session and get records from the json collector file. The file does not have to exist, for
        example when the store is started for a new run.
        """
        try:
            with self.collector_file.open() as f:
                records = json.load(f)
                for record in records:
                    self.append_from_dict(record)
        except FileNotFoundError:
            pass

    def save_local(self):
        self.collector_file.parent.mkdir(parents=True, exist_ok=True)
        with self.collector_file.open("w") as f:
            f.write(self.data_collection.to_json())

    def save(self, force: bool = False) -> None:
        """
        Save the collected data in the AskAnna Backend. Only submit when data is changed or when force is set to True.
        """
        if not self.changed and not force:
            # We will not save when there is nothing changed
            return

        self.save_local()
        if len(self.data_collection.to_dict()) and not self.run_suuid:
            click.echo("The run SUUID is not set for this session. AskAnna cannot submit the data to the platform.")
            click.echo(f"Your {self.collector_type} data is saved locally in:\n  {self.collector_file}")
            return

        self.save_to_askanna()


class MetricCollector(CollectorTemplate):
    collector_type = "metric"
    file_name = "metrics.json"

    metrics = MetricList()

    @property
    def data_collection(self) -> MetricList:
        return self.metrics

    def append_from_dict(self, dict: dict) -> None:
        self.data_collection.append(MetricObject.from_dict(dict))

    def save_to_askanna(self):
        if self.run_suuid:
            RunGateway().metric_update(self.run_suuid, self.metrics)


class VariableCollector(CollectorTemplate):
    collector_type = "variable"
    file_name = "variables.json"

    variables = VariableList()

    @property
    def data_collection(self) -> VariableList:
        return self.variables

    def append_from_dict(self, dict: dict) -> None:
        self.data_collection.append(VariableObject.from_dict(dict))

    def save_to_askanna(self):
        if self.run_suuid:
            RunGateway().variable_update(self.run_suuid, self.variables)



"""This file contains anything that needs to be loaded for `import askanna` to work."""
import click

from askanna import USING_ASKANNA_CLI
from askanna.core.utils.main import update_available
from askanna.sdk.job import JobSDK
from askanna.sdk.project import ProjectSDK
from askanna.sdk.run import ResultSDK, RunSDK
from askanna.sdk.track import (  # noqa: F401
    track_metric,
    track_metrics,
    track_variable,
    track_variables,
)
from askanna.sdk.variable import VariableSDK
from askanna.sdk.workspace import WorkspaceSDK

if USING_ASKANNA_CLI:  # pragma: no cover
    try:
        update_available()
    except Exception as e:
        click.echo(f"Something went wrong while checking if an update is available: {e}", err=True)


# Instantiated objects for query or actions, these do not contain any data on load and will be filled when needed
job = JobSDK()
project = ProjectSDK()
result = ResultSDK()
run = RunSDK()
variable = VariableSDK()
workspace = WorkspaceSDK()



from pathlib import Path

DEFAULT_SERVER_CONFIG_PATH = str(Path("~/.askanna.yml").expanduser())
DEFAULT_SERVER_REMOTE = "https://beta-api.askanna.eu"
DEFAULT_SERVER_UI = "https://beta.askanna.eu"

DEFAULT_PROJECT_TEMPLATE = "https://gitlab.com/askanna/project-templates/blanco-template.git"

PYPI_PROJECT_URL = "https://pypi.org/pypi/askanna/json"



"""
The AskAnna CLI & Python SDK is part of the AskAnna platform to kickstart your data science projects.
"""

__version__ = "0.23.0"

import re
import sys

ASKANNA_VERSION = __version__

# Determine whether we are in the CLI or using the SDK by checking for the `askanna` command
USING_ASKANNA_CLI: bool = any([re.match(".+bin/askanna$", sys.argv[0])])

try:
    import click  # noqa: F401
except ModuleNotFoundError as e:  # pragma: no cover
    # We are propably within an installation for hatch or pip, skip the rest of the initialization
    print(e)
else:
    # We use dotenv for development but it's not a requirement for the CLI or SDK
    try:
        from dotenv import find_dotenv, load_dotenv
    except ImportError:  # pragma: no cover
        pass
    else:  # pragma: no cover
        load_dotenv(find_dotenv())

    # Initialize AskAnna after everything is loaded successfully
    from .init import *  # noqa: F403



from pathlib import Path
from typing import List, Optional, Union

from askanna.core.dataclasses.job import Payload
from askanna.core.dataclasses.run import (
    STATUS,
    TRIGGER,
    ArtifactInfo,
    MetricList,
    MetricObject,
    Run,
    RunStatus,
    VariableList,
    VariableObject,
)
from askanna.core.download import ChunkedDownload
from askanna.core.exceptions import (
    DeleteError,
    GetError,
    HeadError,
    PatchError,
    PutError,
)
from askanna.gateways.api_client import client

from .utils import ListResponse


class RunListResponse(ListResponse):
    def __init__(self, data: dict):
        super().__init__(data)
        self.results: List[Run] = [Run.from_dict(run) for run in data["results"]]

    @property
    def runs(self):
        return self.results


class RunGateway:
    """Management of runs in AskAnna
    This is the class which act as the gateway to the API of AskAnna
    """

    def list(
        self,
        status: Optional[STATUS] = None,
        status__exclude: Optional[STATUS] = None,
        run_suuid_list: Optional[List[str]] = None,
        run_suuid__exclude: Optional[str] = None,
        job_suuid: Optional[str] = None,
        job_suuid__exclude: Optional[str] = None,
        project_suuid: Optional[str] = None,
        project_suuid__exclude: Optional[str] = None,
        workspace_suuid: Optional[str] = None,
        workspace_suuid__exclude: Optional[str] = None,
        created_by_suuid: Optional[str] = None,
        created_by_suuid__exclude: Optional[str] = None,
        trigger: Optional[Union[TRIGGER, List[TRIGGER]]] = None,
        trigger__exclude: Optional[Union[TRIGGER, List[TRIGGER]]] = None,
        package_suuid: Optional[str] = None,
        package_suuid__exclude: Optional[str] = None,
        page_size: Optional[int] = None,
        cursor: Optional[str] = None,
        order_by: Optional[str] = None,
        search: Optional[str] = None,
    ) -> RunListResponse:
        """List all runs with filter and order options

        Args:
            status (STATUS, optional): Status of the run to filter on. Defaults to None.
            status__exclude (str, optional): Status of the run to exclude. Defaults to None.
              STATUS values: queued, running, finished, failed

            run_suuid_list (List[str], optional): List of run SUUIDs to filter on. Defaults to None.
            run_suuid__exclude (str, optional): SUUID of the run to exclude. Defaults to None.

            job_name (str, optional): Name of the job to filter on. Defaults to None.
            job_suuid (str, optional): SUUID of the job to filter on. Defaults to None.
            job_suuid__exclude (str, optional): SUUID of the job to exclude. Defaults to None.

            project_suuid (str, optional): SUUID of the project to filter on. Defaults to None.
            project_suuid__exclude (str, optional): SUUID of the project to exclude. Defaults to None.

            workspace_suuid (str, optional): SUUID of the workspace to filter on. Defaults to None.
            workspace_suuid__exclude (str, optional): SUUID of the workspace to exclude. Defaults to None.

            created_by_suuid (str, optional): SUUID of the workspace member to filter on. Defaults to None.
            created_by_suuid__exclude (str, optional): SUUID of the workspace member to exclude.
              Defaults to None.

            trigger (TRIGGER, optional): Trigger of the run to filter on. Defaults to None.
            trigger__exclude (TRIGGER, optional): Trigger of the run to exclude. Defaults to None.
              TRIGGER values: api, cli, python-sdk, webui, schedule, worker

            package_suuid (str, optional): SUUID of the package to filter on. Defaults to None.
            package_suuid__exclude (str, optional): SUUID of the package to exclude. Defaults to None.

            page_size (int, optional): Number of results per page. Defaults to the default value of the backend.
            cursor (str, optional): Cursor to start the page from. Defaults to None.
            order_by (str, optional): Order by a field. Defaults to None.

            search (str, optional): Search for a specific run. Defaults to None.

        Raises:
            GetError: Error based on response status code with the error message from the API

        Returns:
            RunListResponse: The response from the API with a list of runs and pagination information
        """
        if page_size is not None and (
            (isinstance(page_size, int) and page_size <= 0)
            or isinstance(page_size, bool)
            or not isinstance(page_size, int)
        ):
            raise ValueError("page_size must be a positive integer")

        run_suuid = None
        if run_suuid_list and len(run_suuid_list) > 0:
            run_suuid = ",".join(run_suuid_list)

        response = client.get(
            url=client.askanna_url.run.run_list(),
            params={
                "status": status,
                "status__exclude": status__exclude,
                "run_suuid": run_suuid,
                "run_suuid__exclude": run_suuid__exclude,
                "job_suuid": job_suuid,
                "job_suuid__exclude": job_suuid__exclude,
                "project_suuid": project_suuid,
                "project_suuid__exclude": project_suuid__exclude,
                "workspace_suuid": workspace_suuid,
                "workspace_suuid__exclude": workspace_suuid__exclude,
                "created_by_suuid": created_by_suuid,
                "created_by_suuid__exclude": created_by_suuid__exclude,
                "trigger": trigger,
                "trigger__exclude": trigger__exclude,
                "package_suuid": package_suuid,
                "package_suuid__exclude": package_suuid__exclude,
                "page_size": page_size,
                "cursor": cursor,
                "order_by": order_by,
                "search": search,
            },
        )

        if response.status_code != 200:
            error_message = f"{response.status_code} - Something went wrong while retrieving the run list"
            try:
                error_message += f":\n  {response.json()}"
            except ValueError:
                pass
            raise GetError(error_message)

        return RunListResponse(response.json())

    def detail(self, run_suuid: str) -> Run:
        """Get information of a run

        Args:
            suuid (str): SUUID of the run

        Raises:
            GetError: Error based on response status code with the error message from the API

        Returns:
            Run: Run information in a Run dataclass
        """
        url = client.askanna_url.run.run_detail(run_suuid)
        response = client.get(url)

        if response.status_code == 404:
            raise GetError(f"404 - The run SUUID '{run_suuid}' was not found")
        if response.status_code != 200:
            raise GetError(
                f"{response.status_code} - Something went wrong while retrieving run SUUID '{run_suuid}': "
                f"{response.json()}"
            )

        return Run.from_dict(response.json())

    def change(self, run_suuid: str, name: Optional[str] = None, description: Optional[str] = None) -> Run:
        """Change the name and/or description of a run

        Args:
            run_suuid (str): SUUID of the run you want to change the information of
            name (str, optional): New name for the run. Defaults to None.
            description (str, optional): New description of the run. Defaults to None.

        Raises:
            ValueError: Error if none of the arguments 'name' or 'description' are provided
            PatchError: Error based on response status code with the error message from the API

        Returns:
            Run: The changed run information in a Run dataclass
        """
        changes = {}
        if name:
            changes.update({"name": name})
        if description:
            changes.update({"description": description})

        if not changes:
            raise ValueError("At least one of the parameters 'name' or 'description' must be set.")

        url = client.askanna_url.run.run_detail(run_suuid)
        response = client.patch(url, json=changes)

        if response.status_code == 404:
            raise PatchError(f"404 - The run SUUID '{run_suuid}' was not found")
        if response.status_code != 200:
            raise PatchError(
                f"{response.status_code} - Something went wrong while updating the run SUUID '{run_suuid}': "
                f"{response.json()}"
            )

        return Run.from_dict(response.json())

    def delete(self, run_suuid: str) -> bool:
        """Delete a run

        Args:
            run_suuid (str): SUUID of the run you want to delete

        Raises:
            DeleteError: Error based on response status code with the error message from the API

        Returns:
            bool: True if the run was succesfully deleted
        """
        response = client.delete(
            url=client.askanna_url.run.run_detail(run_suuid),
        )

        if response.status_code == 404:
            raise DeleteError(f"404 - The run SUUID '{run_suuid}' was not found")
        if response.status_code != 204:
            raise DeleteError(
                f"{response.status_code} - Something went wrong while deleting the run SUUID '{run_suuid}': "
                f"{response.json()}"
            )

        return True

    def status(self, run_suuid: str) -> RunStatus:
        """Get the status of a run

        Args:
            run_suuid (str): SUUID of the run you want to get the status of

        Raises:
            GetError: Error based on response status code with the error message from the API

        Returns:
            RunStatus: The status of the run in a RunStatus dataclass
        """
        response = client.get(
            url=client.askanna_url.run.status(run_suuid),
        )

        if response.status_code == 404:
            raise GetError(f"404 - The run SUUID '{run_suuid}' was not found")
        if response.status_code != 200:
            raise GetError(
                f"{response.status_code} - Something went wrong while retrieving the status of run SUUID "
                f"'{run_suuid}': {response.json()}"
            )

        return RunStatus.from_dict(response.json())

    def manifest(
        self,
        run_suuid: str,
        output_path: Optional[Union[Path, str]] = None,
        overwrite: bool = False,
    ) -> Union[bytes, None]:
        """Get the manifest of a run and optionally save it to a file

        Args:
            run_suuid (str): SUUID of the run you want to get the manifest of
            output_path (Path | str, optional): Path to save the manifest to. Defaults to None.
            overwrite (bool, optional): Overwrite a file if it already exists. Defaults to False.

        Raises:
            GetError: Error based on response status code with the error message from the API

        Returns:
            bytes: The manifest of the run
        """
        url = client.askanna_url.run.manifest(run_suuid)

        if output_path:
            download = ChunkedDownload(url)

            if download.status_code == 404:
                raise GetError(f"404 - The manifest for run SUUID '{run_suuid}' was not found")
            if download.status_code != 200:
                raise GetError(
                    f"{download.status_code} - Something went wrong while retrieving the manifest for run SUUID "
                    f"'{run_suuid}'"
                )

            download.download(output_file=output_path, overwrite=overwrite)
            return None

        else:
            response = client.get(url)

            if response.status_code == 404:
                raise GetError(f"404 - The manifest for run SUUID '{run_suuid}' was not found")
            if response.status_code != 200:
                raise GetError(
                    f"{response.status_code} - Something went wrong while retrieving the manifest for run SUUID "
                    f"'{run_suuid}': {response.json()}"
                )

            return response.content

    def metric(self, run_suuid: str) -> MetricList:
        """Get the metrics of a run

        Args:
            run_suuid (str): SUUID of the run you want to get the metrics of

        Raises:
            GetError: Error based on response status code with the error message from the API

        Returns:
            MetricList: The metrics of the run in a MetricList dataclass
        """
        url = client.askanna_url.run.metric(run_suuid)
        response = client.get(url)

        if response.status_code == 404:
            raise GetError(f"404 - The run SUUID '{run_suuid}' was not found")
        if response.status_code != 200:
            raise GetError(
                f"{response.status_code} - Something went wrong while retrieving the metrics of run SUUID "
                f"'{run_suuid}': {response.json()}"
            )

        return MetricList(metrics=[MetricObject.from_dict(metric) for metric in response.json()["results"]])

    def metric_update(self, run_suuid: str, metrics: MetricList) -> None:
        """Update the metrics of a run

        Args:
            run_suuid (str): SUUID of the run you want to update the metrics of
            metrics (MetricList): The list of metrics you want to save

        Raises:
            PutError: Error based on response status code with the error message from the API
        """
        url = client.askanna_url.run.metric_detail(run_suuid)
        response = client.put(url, json={"metrics": metrics.to_dict()})

        if response.status_code == 404:
            raise PutError(f"404 - The run SUUID '{run_suuid}' was not found")
        if response.status_code != 200:
            raise PutError(
                f"{response.status_code} - Something went wrong while updating metrics of run SUUID '{run_suuid}': "
                f"{response.json()}"
            )

    def variable(self, run_suuid: str) -> VariableList:
        """Get the variables of a run

        Args:
            run_suuid (str): SUUID of the run you want to get the variables of

        Raises:
            GetError: Error based on response status code with the error message from the API

        Returns:
            VariableList: The variables of the run in a VariableList dataclass
        """
        url = client.askanna_url.run.variable(run_suuid)
        response = client.get(url)

        if response.status_code == 404:
            raise GetError(f"404 - The run SUUID '{run_suuid}' was not found")
        if response.status_code != 200:
            raise GetError(
                f"{response.status_code} - Something went wrong while retrieving the variables of run SUUID "
                f"'{run_suuid}': {response.json()}"
            )

        return VariableList(variables=[VariableObject.from_dict(variable) for variable in response.json()["results"]])

    def variable_update(self, run_suuid: str, variables: VariableList) -> None:
        """Update the variables of a run

        Args:
            run_suuid (str): SUUID of the run you want to update the variables of
            variables (VariableList): The list of variables you want to save

        Raises:
            PatchError: Error based on response status code with the error message from the API
        """
        url = client.askanna_url.run.variable_detail(run_suuid)
        response = client.patch(url, json={"variables": variables.to_dict()})

        if response.status_code == 404:
            raise PatchError(f"404 - The run SUUID '{run_suuid}' was not found")
        if response.status_code != 200:
            raise PatchError(
                f"{response.status_code} - Something went wrong while updating variables of run SUUID '{run_suuid}': "
                f"{response.json()}"
            )

    def log(self, run_suuid: str, limit: Optional[int] = -1, offset: Optional[int] = None) -> List:
        """Get the log of a run

        Args:
            run_suuid (str): SUUID of the run you want to get the log of.
            limit (int, optional): Limit the number of log lines. Defaults to -1 (all log lines).
            offset (int, optional): Offset the number of log lines. Defaults to None.

        Raises:
            GetError: Error based on response status code with the error message from the API.

        Returns:
            log (List): The log of the run in a list. Each record is a list with 3 items: index, datetime and log line.
        """
        response = client.get(
            url=client.askanna_url.run.log(run_suuid),
            params={
                "limit": limit,
                "offset": offset,
            },
        )

        if response.status_code == 404:
            raise GetError(f"404 - The run SUUID '{run_suuid}' was not found")
        if response.status_code != 200:
            raise GetError(
                f"{response.status_code} - Something went wrong while retrieving the log of run SUUID '{run_suuid}': "
                f"{response.json()}"
            )

        return list(response.json().get("results", []))

    def payload_info(self, run_suuid: str) -> Union[Payload, None]:
        """Get the payload info of a run

        Args:
            run_suuid (str): SUUID of the run you want to get the payload info of

        Raises:
            GetError: Error based on response status code with the error message from the API

        Returns:
            Payload or None: The payload info of the run in a Payload dataclass, or None in case there is no payload
        """
        url = client.askanna_url.run.payload_list(run_suuid)
        response = client.get(url)

        if response.status_code == 404:
            raise GetError(f"404 - The run with run SUUID '{run_suuid}' was not found")
        elif response.status_code != 200:
            raise GetError(
                f"{response.status_code} - Something went wrong while retrieving the payload list of run SUUID "
                f"'{run_suuid}': {response.json()}"
            )

        # The payload list is a list of dicts, but we only want the first one because there could only be one payload
        # per run.
        if len(response.json()) > 0:
            return Payload.from_dict(response.json()[0])
        else:
            return None

    def payload(
        self, run_suuid: str, payload_suuid: str, output_path: Optional[Union[Path, str]] = None
    ) -> Union[bytes, None]:
        """Get the payload of a run and optionally save it to a file

        Args:
            run_suuid (str): SUUID of the run you want to get the payload content of
            payload_suuid (str): SUUID of the payload you want to get the content of
            output_path (Path | str, optional): Path to save the payload to. Defaults to None.

        Raises:
            GetError: Error based on response status code with the error message from the API

        Returns:
            bytes or None: The payload of the run in bytes, or None if output_path is set
        """
        url = client.askanna_url.run.payload_download(run_suuid, payload_suuid)

        if output_path:
            download = ChunkedDownload(url)

            if download.status_code == 404:
                raise GetError(f"404 - The payload for run SUUID '{run_suuid}' was not found")
            if download.status_code != 200:
                raise GetError(
                    f"{download.status_code} - Something went wrong while retrieving the payload for run SUUID "
                    f"'{run_suuid}'"
                )

            download.download(output_path)
            return None

        else:
            response = client.get(url)

            if response.status_code == 404:
                raise GetError(f"404 - The payload for run SUUID '{run_suuid}' was not found")
            if response.status_code != 200:
                raise GetError(
                    f"{response.status_code} - Something went wrong while retrieving the payload for run SUUID "
                    f"'{run_suuid}': {response.json()}"
                )

            return response.content

    def result(self, run_suuid: str, output_path: Optional[Union[Path, str]] = None) -> Union[bytes, None]:
        """Get the result of a run and optionally save it to a file

        Args:
            run_suuid (str): SUUID of the run you want to get the result of
            output_path (Path | str, optional): Path to save the result to. Defaults to None.

        Raises:
            GetError: Error based on response status code with the error message from the API

        Returns:
            bytes or None: The result of the run in bytes, or None if output_path is set
        """
        url = client.askanna_url.run.result(run_suuid)

        if output_path:
            download = ChunkedDownload(url)

            if download.status_code == 404:
                raise GetError(f"404 - The result for run SUUID '{run_suuid}' was not found")
            if download.status_code != 200:
                raise GetError(
                    f"{download.status_code} - Something went wrong while retrieving the result for run SUUID "
                    f"'{run_suuid}'"
                )

            download.download(output_path)
            return None

        else:
            response = client.get(url)

            if response.status_code == 404:
                raise GetError(f"404 - The result for run SUUID '{run_suuid}' was not found")
            if response.status_code != 200:
                raise GetError(
                    f"{response.status_code} - Something went wrong while retrieving the result for run SUUID "
                    f"'{run_suuid}': {response.json()}"
                )

            return response.content

    def result_content_type(self, run_suuid: str) -> str:
        """Get the content type of the result of a run

        Args:
            run_suuid (str): SUUID of the run you want to get the result content type of

        Raises:
            HeadError: Error based on response status code with the error message from the API

        Returns:
            str: The content type of the result of the run
        """

        url = client.askanna_url.run.result(run_suuid)
        response = client.head(url)

        if response.status_code == 404:
            raise HeadError(f"404 - The result for run SUUID '{run_suuid}' was not found")
        if response.status_code != 200:
            raise HeadError(
                f"{response.status_code} - Something went wrong while retrieving the result content type for run "
                f"SUUID '{run_suuid}': {response.json()}"
            )

        return str(response.headers.get("Content-Type"))

    def artifact(
        self,
        run_suuid: str,
        artifact_suuid: Optional[str] = None,
        output_path: Optional[Union[Path, str]] = None,
    ) -> Union[bytes, None]:
        """Get the artifact of a run and optionally save it to a file

        Args:
            run_suuid (str): SUUID of the run you want to get the artifact of
            output_path (Path | str, optional): Path to save the artifact to. Defaults to None.

        Raises:
            GetError: Error based on response status code with the error message from the API

        Returns:
            bytes or None: The artifact of the run in bytes, or None if output_path is set
        """
        artifact_suuid = artifact_suuid or self._get_artifact_suuid(run_suuid)
        url = client.askanna_url.run.artifact_download(run_suuid, artifact_suuid)

        response = client.get(url)

        if response.status_code == 404:
            raise GetError(f"404 - The artifact for run SUUID '{run_suuid}' was not found")
        if response.status_code != 200:
            raise GetError(
                f"{response.status_code} - Something went wrong while retrieving the artifact for run SUUID "
                f"'{run_suuid}': {response.json()}"
            )

        download_url = response.json().get("target")

        if output_path:
            download = ChunkedDownload(download_url)

            if download.status_code == 404:
                raise GetError(f"404 - The artifact for run SUUID '{run_suuid}' was not found")
            if download.status_code != 200:
                raise GetError(
                    f"{download.status_code} - Something went wrong while retrieving the artifact for run SUUID "
                    f"'{run_suuid}'"
                )

            download.download(output_path)
            return None

        else:
            response = client.get(download_url)

            if response.status_code == 404:
                raise GetError(f"404 - The artifact for run SUUID '{run_suuid}' was not found")
            if response.status_code != 200:
                raise GetError(
                    f"{response.status_code} - Something went wrong while retrieving the artifact for run SUUID "
                    f"'{run_suuid}': {response.json()}"
                )

            return response.content

    def artifact_info(self, run_suuid: str, artifact_suuid: Optional[str] = None) -> ArtifactInfo:
        """Get artifact info of a run

        Args:
            run_suuid (str): SUUID of the run you want to get the artifact of
            artifact_suuid (str, optional): SUUID of the artifact you want to get the info of. Defaults to None.

        Raises:
            GetError: Error based on response status code with the error message from the API

        Returns:
            ArtifactInfo: Info about the artifact of a run in a ArtifactInfo dataclass
        """
        artifact_suuid = artifact_suuid or self._get_artifact_suuid(run_suuid)
        url = client.askanna_url.run.artifact_detail(run_suuid, artifact_suuid)

        response = client.get(url)

        if response.status_code == 404:
            raise GetError(f"404 - The artifact for run SUUID '{run_suuid}' was not found")
        if response.status_code != 200:
            raise GetError(
                f"{response.status_code} - Something went wrong while retrieving the artifact for run SUUID "
                f"'{run_suuid}': {response.json()}"
            )

        return ArtifactInfo.from_dict(response.json())

    def _get_artifact_suuid(self, run_suuid: str) -> str:
        """Get the artifact SUUID of the artifact of a run

        Args:
            run_suuid (str): SUUID of the run you want to get the artifact SUUID of

        Raises:
            GetError: Error based on response status code with the error message from the API

        Returns:
            str: SUUID of the artifact of the run
        """
        url = client.askanna_url.run.artifact_list(run_suuid)
        response = client.get(url)

        if response.status_code == 404 or len(response.json()) == 0:
            raise GetError(f"404 - The artifact for run SUUID '{run_suuid}' was not found")
        if response.status_code != 200:
            raise GetError(
                f"{response.status_code} - Something went wrong while retrieving the artifact for run SUUID "
                f"'{run_suuid}': {response.json()}"
            )

        return response.json()[0]["suuid"]
